import json
import os
import platform
import re
import textwrap

import pytest

from conan.tools.microsoft.visual import vcvars_command
from conan.test.assets.cmake import gen_cmakelists
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient
from conan.internal.util.files import save
from test.conftest import tools_locations


@pytest.mark.skipif(platform.system() != "Windows", reason="Only for windows")
@pytest.mark.parametrize("compiler, version, update, runtime",
                         [("msvc", "192", None, "dynamic"),
                          ("msvc", "192", "6", "static"),
                          ("msvc", "192", "8", "static")])
def test_cmake_toolchain_win_toolset(compiler, version, update, runtime):
    client = TestClient(path_with_spaces=False)
    settings = {"compiler": compiler,
                "compiler.version": version,
                "compiler.update": update,
                "compiler.cppstd": "17",
                "compiler.runtime": runtime,
                "build_type": "Release",
                "arch": "x86_64"}

    # Build the profile according to the settings provided
    settings = " ".join('-s %s="%s"' % (k, v) for k, v in settings.items() if v)

    conanfile = GenConanfile().with_settings("os", "compiler", "build_type", "arch").\
        with_generator("CMakeToolchain")

    client.save({"conanfile.py": conanfile})
    client.run("install . {}".format(settings))
    toolchain = client.load("conan_toolchain.cmake")
    value = "v14{}".format(version[-1])
    if update is not None:  # Fullversion
        value += f",version=14.{version[-1]}{update}"
    assert 'set(CMAKE_GENERATOR_TOOLSET "{}" CACHE STRING "" FORCE)'.format(value) in toolchain


def test_cmake_toolchain_user_toolchain():
    client = TestClient(path_with_spaces=False)
    conanfile = GenConanfile().with_settings("os", "compiler", "build_type", "arch").\
        with_generator("CMakeToolchain")
    client.save_home({"global.conf": "tools.cmake.cmaketoolchain:user_toolchain+=mytoolchain.cmake"})

    client.save({"conanfile.py": conanfile})
    client.run("install .")
    toolchain = client.load("conan_toolchain.cmake")
    assert 'include("mytoolchain.cmake")' in toolchain


def test_cmake_toolchain_user_toolchain_from_dep():
    client = TestClient()
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import copy
        class Pkg(ConanFile):
            exports_sources = "*"
            def package(self):
                copy(self, "*", self.build_folder, self.package_folder)
            def package_info(self):
                f = os.path.join(self.package_folder, "mytoolchain.cmake")
                self.conf_info.append("tools.cmake.cmaketoolchain:user_toolchain", f)
        """)
    client.save({"conanfile.py": conanfile,
                 "mytoolchain.cmake": 'message(STATUS "mytoolchain.cmake !!!running!!!")'})
    client.run("create . --name=toolchain --version=0.1")

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake
        class Pkg(ConanFile):
            settings = "os", "compiler", "arch", "build_type"
            exports_sources = "CMakeLists.txt"
            build_requires = "toolchain/0.1"
            generators = "CMakeToolchain"
            def build(self):
                cmake = CMake(self)
                cmake.configure()
        """)

    client.save({"conanfile.py": conanfile,
                 "CMakeLists.txt": gen_cmakelists()}, clean_first=True)
    client.run("create . --name=pkg --version=0.1")
    assert "mytoolchain.cmake !!!running!!!" in client.out


def test_cmake_toolchain_without_build_type():
    # If "build_type" is not defined, toolchain will still be generated, it will not crash
    # Main effect is CMAKE_MSVC_RUNTIME_LIBRARY not being defined
    client = TestClient(path_with_spaces=False)
    conanfile = GenConanfile().with_settings("os", "compiler", "arch").\
        with_generator("CMakeToolchain")

    client.save({"conanfile.py": conanfile})
    client.run("install .")
    toolchain = client.load("conan_toolchain.cmake")
    assert "CMAKE_MSVC_RUNTIME_LIBRARY" not in toolchain
    assert "CMAKE_BUILD_TYPE" not in toolchain


@pytest.mark.skipif(platform.system() != "Windows", reason="Only on Windows with msvc")
@pytest.mark.tool("cmake")
def test_cmake_toolchain_cmake_vs_debugger_environment():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile("pkg", "1.0").with_package_type("shared-library")
                                                           .with_settings("build_type")})
    client.run("create . -s build_type=Release")
    client.run("create . -s build_type=Debug")
    client.run("create . -s build_type=MinSizeRel")

    client.run("install --require=pkg/1.0 -s build_type=Debug -g CMakeToolchain --format=json")
    debug_graph = json.loads(client.stdout)
    debug_bindir = debug_graph['graph']['nodes']['1']['cpp_info']['root']['bindirs'][0]
    debug_bindir = debug_bindir.replace('\\', '/')

    toolchain = client.load("conan_toolchain.cmake")
    debugger_environment = f"PATH=$<$<CONFIG:Debug>:{debug_bindir}>;%PATH%"
    assert debugger_environment in toolchain

    client.run("install --require=pkg/1.0 -s build_type=Release -g CMakeToolchain --format=json")
    release_graph = json.loads(client.stdout)
    release_bindir = release_graph['graph']['nodes']['1']['cpp_info']['root']['bindirs'][0]
    release_bindir = release_bindir.replace('\\', '/')

    toolchain = client.load("conan_toolchain.cmake")
    debugger_environment = f"PATH=$<$<CONFIG:Debug>:{debug_bindir}>" \
                           f"$<$<CONFIG:Release>:{release_bindir}>;%PATH%"
    assert debugger_environment in toolchain

    client.run("install --require=pkg/1.0 -s build_type=MinSizeRel -g CMakeToolchain --format=json")
    minsizerel_graph = json.loads(client.stdout)
    minsizerel_bindir = minsizerel_graph['graph']['nodes']['1']['cpp_info']['root']['bindirs'][0]
    minsizerel_bindir = minsizerel_bindir.replace('\\', '/')

    toolchain = client.load("conan_toolchain.cmake")
    debugger_environment = f"PATH=$<$<CONFIG:Debug>:{debug_bindir}>" \
                           f"$<$<CONFIG:Release>:{release_bindir}>" \
                           f"$<$<CONFIG:MinSizeRel>:{minsizerel_bindir}>;%PATH%"
    assert debugger_environment in toolchain


@pytest.mark.tool("cmake")
def test_cmake_toolchain_cmake_vs_debugger_environment_not_needed():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile("pkg", "1.0").with_package_type("shared-library")
                                                           .with_settings("build_type")})
    client.run("create . -s build_type=Release")

    cmake_generator = "" if platform.system() != "Windows" else "-c tools.cmake.cmaketoolchain:generator=Ninja"
    client.run(f"install --require=pkg/1.0 -s build_type=Release -g CMakeToolchain {cmake_generator}")
    toolchain = client.load("conan_toolchain.cmake")
    assert "CMAKE_VS_DEBUGGER_ENVIRONMENT" not in toolchain


@pytest.mark.tool("cmake")
def test_cmake_toolchain_multiple_user_toolchain():
    """ A consumer consuming two packages that declare:
            self.conf_info["tools.cmake.cmaketoolchain:user_toolchain"]
        The consumer wants to use apply both toolchains in the CMakeToolchain.
        There are two ways to customize the CMakeToolchain (parametrized):
                1. Altering the context of the block (with_context = True)
                2. Using the t.blocks["user_toolchain"].user_toolchains = [] (with_context = False)
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import copy
        class Pkg(ConanFile):
            exports_sources = "*"
            def package(self):
                copy(self, "*", self.source_folder, self.package_folder)
            def package_info(self):
                f = os.path.join(self.package_folder, "mytoolchain.cmake")
                self.conf_info.append("tools.cmake.cmaketoolchain:user_toolchain", f)
        """)
    client.save({"conanfile.py": conanfile,
                 "mytoolchain.cmake": 'message(STATUS "mytoolchain1.cmake !!!running!!!")'})
    client.run("create . --name=toolchain1 --version=0.1")
    client.save({"conanfile.py": conanfile,
                 "mytoolchain.cmake": 'message(STATUS "mytoolchain2.cmake !!!running!!!")'})
    client.run("create . --name=toolchain2 --version=0.1")

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake, CMakeToolchain
        class Pkg(ConanFile):
            settings = "os", "compiler", "arch", "build_type"
            exports_sources = "CMakeLists.txt"
            tool_requires = "toolchain1/0.1", "toolchain2/0.1"
            generators = "CMakeToolchain"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
        """)

    client.save({"conanfile.py": conanfile,
                 "CMakeLists.txt": gen_cmakelists()}, clean_first=True)
    client.run("create . --name=pkg --version=0.1")
    assert "mytoolchain1.cmake !!!running!!!" in client.out
    assert "mytoolchain2.cmake !!!running!!!" in client.out
    assert "CMake Warning" not in client.out


@pytest.mark.tool("cmake")
def test_cmaketoolchain_no_warnings():
    """Make sure uninitialized variables do not cause any warnings, passing -Werror=dev
    and --warn-uninitialized, calling "cmake" with conan_toolchain.cmake used to fail
    """
    # Issue https://github.com/conan-io/conan/issues/10288
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Conan(ConanFile):
            settings = "os", "compiler", "arch", "build_type"
            generators = "CMakeToolchain", "CMakeDeps"
            requires = "dep/0.1"
        """)
    consumer = textwrap.dedent("""
       cmake_minimum_required(VERSION 3.15)
       project(MyHello NONE)

       find_package(dep CONFIG REQUIRED)
       """)
    client.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
                 "conanfile.py": conanfile,
                 "CMakeLists.txt": consumer})

    client.run("create dep")
    client.run("install .")
    build_type = "-DCMAKE_BUILD_TYPE=Release" if platform.system() != "Windows" else ""
    client.run_command("cmake -Werror=dev --warn-uninitialized . {}"
                       " -DCMAKE_TOOLCHAIN_FILE=./conan_toolchain.cmake".format(build_type))
    assert "Using Conan toolchain" in client.out
    # The real test is that there are no errors, it returns successfully


def test_install_output_directories():
    """
    If we change the libdirs of the cpp.package, as we are doing cmake.install, the output directory
    for the libraries is changed
    """
    client = TestClient()
    client.run("new cmake_lib -d name=zlib -d version=1.2.11")
    # Edit the cpp.package.libdirs and check if the library is placed anywhere else
    cf = client.load("conanfile.py")
    cf = cf.replace("cmake_layout(self)",
                    'cmake_layout(self)\n        self.cpp.package.libdirs = ["mylibs"]')
    client.save({"conanfile.py": cf})
    client.run("create . -tf=")
    layout = client.created_layout()
    p_folder = layout.package()
    assert os.path.exists(os.path.join(p_folder, "mylibs"))
    assert not os.path.exists(os.path.join(p_folder, "lib"))

    b_folder = layout.build()
    if platform.system() != "Windows":
        gen_folder = os.path.join(b_folder, "build", "Release", "generators")
    else:
        gen_folder = os.path.join(b_folder, "build", "generators")

    toolchain = client.load(os.path.join(gen_folder, "conan_toolchain.cmake"))
    assert 'set(CMAKE_INSTALL_LIBDIR "mylibs")' in toolchain


@pytest.mark.tool("cmake")
def test_cmake_toolchain_definitions_complex_strings():
    # https://github.com/conan-io/conan/issues/11043
    client = TestClient(path_with_spaces=False)
    profile = textwrap.dedent(r'''
        include(default)
        [conf]
        tools.build:defines+=["escape=partially \"escaped\""]
        tools.build:defines+=["spaces=me you"]
        tools.build:defines+=["foobar=bazbuz"]
        tools.build:defines+=["answer=42"]
    ''')

    conanfile = textwrap.dedent(r'''
        from conan import ConanFile
        from conan.tools.cmake import CMake, CMakeToolchain, cmake_layout

        class Test(ConanFile):
            exports_sources = "CMakeLists.txt", "src/*"
            settings = "os", "compiler", "arch", "build_type"

            def generate(self):
                tc = CMakeToolchain(self)
                tc.preprocessor_definitions["escape2"] = "partially \"escaped\""
                tc.preprocessor_definitions["spaces2"] = "me you"
                tc.preprocessor_definitions["foobar2"] = "bazbuz"
                tc.preprocessor_definitions["answer2"] = 42
                tc.preprocessor_definitions["NOVALUE_DEF"] = None
                tc.preprocessor_definitions.release["escape_release"] = "release partially \"escaped\""
                tc.preprocessor_definitions.release["spaces_release"] = "release me you"
                tc.preprocessor_definitions.release["foobar_release"] = "release bazbuz"
                tc.preprocessor_definitions.release["answer_release"] = 42
                tc.preprocessor_definitions.release["NOVALUE_DEF_RELEASE"] = None

                tc.preprocessor_definitions.debug["escape_debug"] = "debug partially \"escaped\""
                tc.preprocessor_definitions.debug["spaces_debug"] = "debug me you"
                tc.preprocessor_definitions.debug["foobar_debug"] = "debug bazbuz"
                tc.preprocessor_definitions.debug["answer_debug"] = 21
                tc.generate()

            def layout(self):
                cmake_layout(self)

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
        ''')

    main = textwrap.dedent("""
        #include <stdio.h>
        #define STR(x)   #x
        #define SHOW_DEFINE(x) printf("%s=%s", #x, STR(x))
        int main(int argc, char *argv[]) {
            SHOW_DEFINE(escape);
            SHOW_DEFINE(spaces);
            SHOW_DEFINE(foobar);
            SHOW_DEFINE(answer);
            SHOW_DEFINE(escape2);
            SHOW_DEFINE(spaces2);
            SHOW_DEFINE(foobar2);
            SHOW_DEFINE(answer2);
            #ifdef NDEBUG
            SHOW_DEFINE(escape_release);
            SHOW_DEFINE(spaces_release);
            SHOW_DEFINE(foobar_release);
            SHOW_DEFINE(answer_release);
            #else
            SHOW_DEFINE(escape_debug);
            SHOW_DEFINE(spaces_debug);
            SHOW_DEFINE(foobar_debug);
            SHOW_DEFINE(answer_debug);
            #endif
            #ifdef NOVALUE_DEF
            printf("NO VALUE!!!!");
            #endif
            #ifdef NOVALUE_DEF_RELEASE
            printf("NO VALUE RELEASE!!!!");
            #endif
            return 0;
        }
        """)

    cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.15)
        project(Test CXX)
        set(CMAKE_CXX_STANDARD 11)
        add_executable(example src/main.cpp)
        """)

    client.save({"conanfile.py": conanfile, "profile": profile, "src/main.cpp": main,
                 "CMakeLists.txt": cmakelists}, clean_first=True)
    client.run("install . -pr=./profile")
    client.run("build . -pr=./profile")
    exe = "build/Release/example" if platform.system() != "Windows" else r"build\Release\example.exe"
    client.run_command(exe)
    assert 'escape=partially "escaped"' in client.out
    assert 'spaces=me you' in client.out
    assert 'foobar=bazbuz' in client.out
    assert 'answer=42' in client.out
    assert 'escape2=partially "escaped"' in client.out
    assert 'spaces2=me you' in client.out
    assert 'foobar2=bazbuz' in client.out
    assert 'answer2=42' in client.out
    assert 'escape_release=release partially "escaped"' in client.out
    assert 'spaces_release=release me you' in client.out
    assert 'foobar_release=release bazbuz' in client.out
    assert 'answer_release=42' in client.out
    assert "NO VALUE!!!!" in client.out
    assert "NO VALUE RELEASE!!!!" in client.out

    client.run("install . -pr=./profile -s build_type=Debug")
    client.run("build . -pr=./profile -s build_type=Debug")
    exe = "build/Debug/example" if platform.system() != "Windows" else r"build\Debug\example.exe"

    client.run_command(exe)
    assert 'escape_debug=debug partially "escaped"' in client.out
    assert 'spaces_debug=debug me you' in client.out
    assert 'foobar_debug=debug bazbuz' in client.out
    assert 'answer_debug=21' in client.out


@pytest.mark.skipif(platform.system() != "Windows", reason="Only for windows")
def test_cmake_toolchain_runtime_types():
    # everything works with the default cmake_minimum_required version 3.15 in the template
    client = TestClient(path_with_spaces=False)
    client.run("new cmake_lib -d name=hello -d version=0.1")
    client.run("install . -s compiler.runtime=static -s build_type=Debug")
    client.run("build . -s compiler.runtime=static -s build_type=Debug")

    vcvars = vcvars_command(version="15", architecture="x64")
    lib = os.path.join(client.current_folder, "build", "Debug", "hello.lib")
    dumpbind_cmd = '{} && dumpbin /directives "{}"'.format(vcvars, lib)
    client.run_command(dumpbind_cmd)
    assert "LIBCMTD" in client.out


@pytest.mark.skipif(platform.system() != "Windows", reason="Only for windows")
def test_cmake_toolchain_runtime_types_cmake_older_than_3_15():
    client = TestClient(path_with_spaces=False)
    # Setting an older cmake_minimum_required in the CMakeLists fails, will link
    # against the default debug runtime (MDd->MSVCRTD), not against MTd->LIBCMTD
    client.run("new cmake_lib -d name=hello -d version=0.1")
    cmake = client.load("CMakeLists.txt")
    cmake2 = cmake.replace('cmake_minimum_required(VERSION 3.15)',
                           'cmake_minimum_required(VERSION 3.1)')
    assert cmake != cmake2
    client.save({"CMakeLists.txt": cmake2})

    client.run("install . -s compiler.runtime=static -s build_type=Debug")
    client.run("build . -s compiler.runtime=static -s build_type=Debug")

    vcvars = vcvars_command(version="15", architecture="x64")
    lib = os.path.join(client.current_folder, "build", "Debug", "hello.lib")
    dumpbind_cmd = '{} && dumpbin /directives "{}"'.format(vcvars, lib)
    client.run_command(dumpbind_cmd)
    assert "LIBCMTD" in client.out


@pytest.mark.skipif(platform.system() != "Windows", reason="Only for windows")
class TestWinSDKVersion:

    @pytest.mark.tool("cmake", "3.23")
    @pytest.mark.tool("visual_studio", "17")
    def test_cmake_toolchain_winsdk_version(self):
        # This test passes also with CMake 3.28, as long as cmake_minimum_required(VERSION 3.27)
        # is not defined
        client = TestClient(path_with_spaces=False)
        client.run("new cmake_lib -d name=hello -d version=0.1")
        cmake = client.load("CMakeLists.txt")
        cmake += 'message(STATUS "CMAKE_VS_WINDOWS_TARGET_PLATFORM_VERSION = ' \
                 '${CMAKE_VS_WINDOWS_TARGET_PLATFORM_VERSION}")'
        client.save({"CMakeLists.txt": cmake})
        client.run("create . -s arch=x86_64 -s compiler.version=194 "
                   "-c tools.microsoft:winsdk_version=10.0")
        assert "CMAKE_VS_WINDOWS_TARGET_PLATFORM_VERSION = 10.0" in client.out
        assert "Conan toolchain: CMAKE_GENERATOR_PLATFORM=x64" in client.out
        assert "Conan toolchain: CMAKE_GENERATOR_PLATFORM=x64,version" not in client.out

    @pytest.mark.tool("cmake", "3.27")
    @pytest.mark.tool("visual_studio", "17")
    def test_cmake_toolchain_winsdk_version2(self):
        # https://github.com/conan-io/conan/issues/15372
        client = TestClient(path_with_spaces=False)
        client.run("new cmake_lib -d name=hello -d version=0.1")
        cmake = client.load("CMakeLists.txt")
        cmake = cmake.replace("cmake_minimum_required(VERSION 3.15)",
                              "cmake_minimum_required(VERSION 3.27)")
        cmake += 'message(STATUS "CMAKE_VS_WINDOWS_TARGET_PLATFORM_VERSION = ' \
                 '${CMAKE_VS_WINDOWS_TARGET_PLATFORM_VERSION}")'
        client.save({"CMakeLists.txt": cmake})
        client.run("create . -s arch=x86_64 -s compiler.version=194 "
                   "-c tools.microsoft:winsdk_version=10.0 "
                   '-c tools.cmake.cmaketoolchain:generator="Visual Studio 17"')
        assert "CMAKE_VS_WINDOWS_TARGET_PLATFORM_VERSION = 10.0" in client.out
        assert "Conan toolchain: CMAKE_GENERATOR_PLATFORM=x64,version=10.0" in client.out


@pytest.mark.tool("cmake", "3.23")
def test_cmake_layout_missing_option():
    client = TestClient(path_with_spaces=False)

    client.run("new cmake_exe -d name=hello -d version=0.1")
    settings_layout = '-c tools.cmake.cmake_layout:build_folder_vars=\'["options.missing"]\' ' \
                      '-c tools.cmake.cmaketoolchain:generator=Ninja'
    client.run("install . {}".format(settings_layout))
    assert os.path.exists(os.path.join(client.current_folder, "build", "Release", "generators"))


@pytest.mark.tool("cmake", "3.23")
def test_cmake_layout_missing_setting():
    client = TestClient(path_with_spaces=False)
    client.run("new cmake_exe -d name=hello -d version=0.1")
    settings_layout = '-c tools.cmake.cmake_layout:build_folder_vars=\'["settings.missing"]\' ' \
                      '-c tools.cmake.cmaketoolchain:generator=Ninja'
    client.run("install . {}".format(settings_layout))
    assert os.path.exists(os.path.join(client.current_folder, "build", "Release", "generators"))


@pytest.mark.tool("cmake")
def test_cmaketoolchain_sysroot():
    client = TestClient(path_with_spaces=False)

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMakeToolchain, CMake, cmake_layout

        class AppConan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            exports_sources = "CMakeLists.txt"

            def generate(self):
                tc = CMakeToolchain(self)
                {}
                tc.generate()

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
        """)

    cmakelist = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.15)
        project(app NONE)
        message("sysroot: '${CMAKE_SYSROOT}'")
        message("osx_sysroot: '${CMAKE_OSX_SYSROOT}'")
        """)

    client.save({
        "conanfile.py": conanfile.format(""),
        "CMakeLists.txt": cmakelist
    })

    fake_sysroot = client.current_folder
    output_fake_sysroot = fake_sysroot.replace("\\", "/") if platform.system() == "Windows" else fake_sysroot
    client.run("create . --name=app --version=1.0 -c tools.build:sysroot='{}'".format(fake_sysroot))
    assert "sysroot: '{}'".format(output_fake_sysroot) in client.out

    # set in a block instead of using conf
    set_sysroot_in_block = 'tc.blocks["generic_system"].values["cmake_sysroot"] = "{}"'.format(output_fake_sysroot)
    client.save({
        "conanfile.py": conanfile.format(set_sysroot_in_block),
    })
    client.run("create . --name=app --version=1.0")
    assert "sysroot: '{}'".format(output_fake_sysroot) in client.out


def test_cmake_layout_not_forbidden_build_type():
    client = TestClient(path_with_spaces=False)
    client.run("new cmake_exe -d name=hello -d version=0.1")
    settings_layout = '-c tools.cmake.cmake_layout:build_folder_vars=' \
                      '\'["options.missing", "settings.build_type"]\''
    client.run("install . {}".format(settings_layout))
    assert os.path.exists(os.path.join(client.current_folder,
                                       "build/release/generators/conan_toolchain.cmake"))


def test_resdirs_cmake_install():
    """If resdirs is declared, the CMAKE_INSTALL_DATAROOTDIR folder is set"""

    client = TestClient(path_with_spaces=False)

    conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.cmake import CMakeToolchain, CMake, cmake_layout

            class AppConan(ConanFile):
                settings = "os", "compiler", "build_type", "arch"
                exports_sources = "CMakeLists.txt", "my_license"
                name = "foo"
                version = "1.0"

                def generate(self):
                    tc = CMakeToolchain(self)
                    tc.generate()

                def layout(self):
                    self.cpp.package.resdirs = ["res"]

                def build(self):
                    cmake = CMake(self)
                    cmake.configure()
                    cmake.build()

                def package(self):
                    cmake = CMake(self)
                    cmake.install()
            """)

    cmake = """
    cmake_minimum_required(VERSION 3.15)
    project(foo NONE)
    if(NOT CMAKE_INSTALL_DATAROOTDIR)
        message(FATAL_ERROR "Cannot install stuff")
    endif()
    install(FILES my_license DESTINATION ${CMAKE_INSTALL_DATAROOTDIR}/licenses)
    """

    client.save({"conanfile.py": conanfile, "CMakeLists.txt": cmake, "my_license": "MIT"})
    client.run("create .")
    assert "/res/licenses/my_license" in client.out
    assert "Packaged 1 file: my_license" in client.out


def test_resdirs_none_cmake_install():
    """If no resdirs are declared, the CMAKE_INSTALL_DATAROOTDIR folder is not set"""

    client = TestClient(path_with_spaces=False)

    conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.cmake import CMakeToolchain, CMake, cmake_layout

            class AppConan(ConanFile):
                settings = "os", "compiler", "build_type", "arch"
                exports_sources = "CMakeLists.txt", "my_license"
                name = "foo"
                version = "1.0"

                def generate(self):
                    tc = CMakeToolchain(self)
                    tc.generate()

                def build(self):
                    cmake = CMake(self)
                    cmake.configure()
                    cmake.build()

                def package(self):
                    cmake = CMake(self)
                    cmake.install()
            """)

    cmake = """
    cmake_minimum_required(VERSION 3.15)
    project(foo NONE)
    if(NOT CMAKE_INSTALL_DATAROOTDIR)
        message(FATAL_ERROR "Cannot install stuff")
    endif()
    """

    client.save({"conanfile.py": conanfile, "CMakeLists.txt": cmake, "my_license": "MIT"})
    client.run("create .", assert_error=True)
    assert "Cannot install stuff" in client.out


@pytest.mark.tool("cmake")
def test_cmake_toolchain_vars_when_option_declared():
    t = TestClient()

    cmakelists = textwrap.dedent("""
    cmake_minimum_required(VERSION 2.8) # <---- set this to an old version for old policies
    cmake_policy(SET CMP0091 NEW) # <-- Needed on Windows
    project(mylib CXX)

    message("CMake version: ${CMAKE_VERSION}")

    # Set the options AFTER the call to project, that is, after toolchain is loaded
    option(BUILD_SHARED_LIBS "" ON)
    option(CMAKE_POSITION_INDEPENDENT_CODE "" OFF)

    add_library(mylib src/mylib.cpp)
    target_include_directories(mylib PUBLIC include)

    get_target_property(MYLIB_TARGET_TYPE mylib TYPE)
    get_target_property(MYLIB_PIC mylib POSITION_INDEPENDENT_CODE)
    message("mylib target type: ${MYLIB_TARGET_TYPE}")
    message("MYLIB_PIC value: ${MYLIB_PIC}")
    if(MYLIB_PIC)
        #Note: the value is "True"/"False" if set by cmake,
        #       and "ON"/"OFF" if set by Conan
        message("mylib position independent code: ON")
    else()
        message("mylib position independent code: OFF")
    endif()

    set_target_properties(mylib PROPERTIES PUBLIC_HEADER "include/mylib.h")
    install(TARGETS mylib)
    """)

    t.run("new cmake_lib -d name=mylib -d version=1.0")
    t.save({"CMakeLists.txt": cmakelists})

    # The generated toolchain should set `BUILD_SHARED_LIBS` to `OFF`,
    # and `CMAKE_POSITION_INDEPENDENT_CODE` to `ON` and the calls to
    # `option()` in the CMakeLists.txt should respect the existing values.
    # Note: on *WINDOWS* `fPIC` is not an option for this recipe, so it's invalid
    #       to pass it to Conan, in which case the value in CMakeLists.txt
    #       takes precedence.
    fpic_option = "-o mylib/*:fPIC=True" if platform.system() != "Windows" else ""
    fpic_cmake_value = "ON" if platform.system() != "Windows" else "OFF"
    t.run(f"create . -o mylib/*:shared=False {fpic_option} --test-folder=")
    assert "mylib target type: STATIC_LIBRARY" in t.out
    assert f"mylib position independent code: {fpic_cmake_value}" in t.out

    # When building manually, ensure the value passed by the toolchain overrides the ones in
    # the CMakeLists
    fpic_option = "-o mylib/*:fPIC=False" if platform.system() != "Windows" else ""
    t.run(f"install . -o mylib/*:shared=False {fpic_option}")
    folder = "build/generators" if platform.system() == "Windows" else "build/Release/generators"
    t.run_command(f"cmake -S . -B build/ -DCMAKE_TOOLCHAIN_FILE={folder}/conan_toolchain.cmake")
    assert "mylib target type: STATIC_LIBRARY" in t.out
    assert f"mylib position independent code: OFF" in t.out

    # Note: from this point forward, the CMakeCache is already initialised.
    # When explicitly overriding `CMAKE_POSITION_INDEPENDENT_CODE` via command line, ensure
    # this takes precedence to the value defined by the toolchain
    t.run_command("cmake -S . -B build/ -DCMAKE_POSITION_INDEPENDENT_CODE=ON")
    assert "mylib target type: STATIC_LIBRARY" in t.out
    assert "mylib position independent code: ON" in t.out

    t.run_command("cmake -S . -B build/ -DBUILD_SHARED_LIBS=ON")
    assert "mylib target type: SHARED_LIBRARY" in t.out
    assert "mylib position independent code: ON" in t.out


@pytest.mark.tool("cmake")
@pytest.mark.parametrize("single_profile", [True, False])
def test_find_program_for_tool_requires(single_profile):
    """Test that the same reference can be both a tool_requires and a regular requires,
    and that find_program (executables) and find_package (libraries) find the correct ones
    when cross building.
    """

    client = TestClient()

    conanfile = textwrap.dedent("""
        import os

        from conan import ConanFile
        from conan.tools.files import copy
        class TestConan(ConanFile):
            name = "foobar"
            version = "1.0"
            settings = "os", "arch", "compiler", "build_type"
            exports_sources = "*"
            def layout(self):
                pass
            def package(self):
                copy(self, pattern="lib*", src=self.build_folder, dst=os.path.join(self.package_folder, "lib"))
                copy(self, pattern="*bin", src=self.build_folder, dst=os.path.join(self.package_folder, "bin"))
    """)

    host_profile = textwrap.dedent("""
    [settings]
        os=Linux
        arch=armv8
        compiler=gcc
        compiler.version=12
        compiler.libcxx=libstdc++11
        build_type=Release
    """)

    build_profile = textwrap.dedent("""
        [settings]
        os=Linux
        arch=x86_64
        compiler=gcc
        compiler.version=12
        compiler.libcxx=libstdc++11
        build_type=Release
    """)

    client.save({"conanfile.py": conanfile,
                "libfoo.so": "",
                "foobin": "",
                "host_profile": host_profile,
                "build_profile": build_profile
                })

    client.run("create . -pr:b build_profile -pr:h build_profile")
    build_context_package_folder = re.search(r"Package folder ([\w\W]+).conan2([\w\W]+)", str(client.out)).group(2).strip()
    build_context_package_folder = build_context_package_folder.replace("\\", "/")
    client.run("create . -pr:b build_profile -pr:h host_profile")
    host_context_package_folder = re.search(r"Package folder ([\w\W]+).conan2([\w\W]+)", str(client.out)).group(2).strip()
    host_context_package_folder = host_context_package_folder.replace("\\", "/")

    conanfile_consumer = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout
        class PkgConan(ConanFile):
            settings = "os", "arch", "compiler", "build_type"

            def layout(self):
                cmake_layout(self)

            def requirements(self):
                self.requires("foobar/1.0")

            def build_requirements(self):
               self.tool_requires("foobar/1.0")
    """)

    cmakelists_consumer = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.15)
        project(Hello LANGUAGES NONE)
        find_package(foobar CONFIG REQUIRED)
        find_program(FOOBIN_EXECUTABLE foobin)
        message("foobin executable: ${FOOBIN_EXECUTABLE}")
        message("foobar include dir: ${foobar_INCLUDE_DIR}")
        if(NOT FOOBIN_EXECUTABLE)
          message(FATAL_ERROR "FOOBIN executable not found")
        endif()
    """)

    client.save({
        "conanfile_consumer.py": conanfile_consumer,
        "CMakeLists.txt": cmakelists_consumer,
        "host_profile": host_profile,
        "build_profile": build_profile}, clean_first=True)

    client.run("install conanfile_consumer.py -g CMakeToolchain -g CMakeDeps -pr:b build_profile -pr:h host_profile")

    with client.chdir("build"):
        client.run_command("cmake .. -DCMAKE_TOOLCHAIN_FILE=Release/generators/conan_toolchain.cmake -DCMAKE_BUILD_TYPE=Release")
        # Verify binary executable is found from build context package,
        # and library comes from host context package
        assert f"{build_context_package_folder}/bin/foobin" in client.out
        assert f"{host_context_package_folder}/include" in client.out


@pytest.mark.tool("pkg_config")
def test_cmaketoolchain_and_pkg_config_path():
    """
    Lightweight test which is loading a dependency as a *.pc file through
    CMake thanks to pkg_check_modules macro.
    It's working because PKG_CONFIG_PATH env variable is being loaded automatically
    by the CMakeToolchain generator.
    """
    client = TestClient()
    dep = textwrap.dedent("""
    from conan import ConanFile
    class DepConan(ConanFile):
        name = "dep"
        version = "1.0"
        def package_info(self):
            self.cpp_info.libs = ["dep"]
    """)
    pkg = textwrap.dedent("""
    from conan import ConanFile
    from conan.tools.cmake import CMake, cmake_layout
    class HelloConan(ConanFile):
        name = "pkg"
        version = "1.0"
        settings = "os", "compiler", "build_type", "arch"
        generators = "PkgConfigDeps", "CMakeToolchain"
        exports_sources = "CMakeLists.txt"

        def requirements(self):
            self.requires("dep/1.0")

        def layout(self):
            cmake_layout(self)

        def build(self):
            cmake = CMake(self)
            cmake.configure()
            cmake.build()
    """)
    cmakelists = textwrap.dedent("""
    cmake_minimum_required(VERSION 3.15)
    project(pkg NONE)

    find_package(PkgConfig REQUIRED)
    # We should have PKG_CONFIG_PATH created in the current environment
    pkg_check_modules(DEP REQUIRED IMPORTED_TARGET dep)
    """)
    client.save({
        "dep/conanfile.py": dep,
        "pkg/conanfile.py": pkg,
        "pkg/CMakeLists.txt": cmakelists
    })
    client.run("create dep/conanfile.py")
    client.run("create pkg/conanfile.py")
    assert "Found dep, version 1.0" in client.out


def test_cmaketoolchain_conf_from_tool_require():
    # https://github.com/conan-io/conan/issues/13914
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Conan(ConanFile):
            name = "toolchain"
            version = "1.0"

            def package_info(self):
                self.conf_info.define("tools.cmake.cmaketoolchain:system_name", "GENERIC-POTATO")
                self.conf_info.define("tools.cmake.cmaketoolchain:system_processor", "ARM-POTATO")
                self.conf_info.define("tools.build.cross_building:can_run", False)
                self.conf_info.define("tools.build:compiler_executables", {
                    "c": "arm-none-eabi-gcc",
                    "cpp": "arm-none-eabi-g++",
                    "asm": "arm-none-eabi-as",
                })
        """)
    c.save({"conanfile.py": conanfile,
            "test_package/conanfile.py": GenConanfile().with_test("pass")
                                                       .with_tool_requires("toolchain/1.0")
                                                       .with_generator("CMakeToolchain")})
    c.run("create .")
    toolchain = c.load("test_package/conan_toolchain.cmake")
    assert "set(CMAKE_SYSTEM_NAME GENERIC-POTATO)" in toolchain
    assert "set(CMAKE_SYSTEM_PROCESSOR ARM-POTATO)" in toolchain


def test_inject_user_toolchain():
    client = TestClient()

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake

        class AppConan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            exports_sources = "CMakeLists.txt"
            name = "foo"
            version = "1.0"
            generators = "CMakeToolchain"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
        """)

    cmake = """
        cmake_minimum_required(VERSION 3.15)
        project(foo LANGUAGES NONE)
        message(STATUS "MYVAR1 ${MY_USER_VAR1}!!")
    """
    profile = textwrap.dedent("""
        include(default)
        [conf]
        tools.cmake.cmaketoolchain:user_toolchain+={{profile_dir}}/myvars.cmake""")
    save(os.path.join(client.paths.profiles_path, "myprofile"), profile)
    save(os.path.join(client.paths.profiles_path, "myvars.cmake"), 'set(MY_USER_VAR1 "MYVALUE1")')
    client.save({"conanfile.py": conanfile,
                 "CMakeLists.txt": cmake})
    client.run("build . -pr=myprofile")
    assert "-- MYVAR1 MYVALUE1!!" in client.out

    # Now test with the global.conf
    global_conf = 'tools.cmake.cmaketoolchain:user_toolchain=' \
                  '["{{conan_home_folder}}/my.cmake"]'
    client.save_home({"global.conf": global_conf})
    save(os.path.join(client.cache_folder, "my.cmake"), 'message(STATUS "IT WORKS!!!!")')
    client.run("build .")
    # The toolchain is found and can be used
    assert "IT WORKS!!!!" in client.out


def test_no_build_type():
    client = TestClient()

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake, CMakeToolchain

        class AppConan(ConanFile):
            name = "pkg"
            version = "1.0"
            settings = "os", "compiler", "arch"
            exports_sources = "CMakeLists.txt"

            def layout(self):
                self.folders.build = "build"

            def generate(self):
                tc = CMakeToolchain(self)
                if not tc.is_multi_configuration:
                    tc.cache_variables["CMAKE_BUILD_TYPE"] = "Release"
                tc.generate()

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                build_type = "Release" if cmake.is_multi_configuration else None
                cmake.build(build_type=build_type)

            def package(self):
                cmake = CMake(self)
                build_type = "Release" if cmake.is_multi_configuration else None
                cmake.install(build_type=build_type)
        """)

    cmake = """
        cmake_minimum_required(VERSION 3.15)
        project(pkg LANGUAGES NONE)
    """

    client.save({"conanfile.py": conanfile,
                 "CMakeLists.txt": cmake})
    client.run("create .")
    assert "Don't specify 'build_type' at build time" not in client.out


@pytest.mark.tool("cmake", "3.19")
def test_redirect_stdout():
    client = TestClient()
    conanfile = textwrap.dedent("""
    import os
    from io import StringIO
    from conan import ConanFile
    from conan.tools.cmake import CMake, CMakeToolchain
    from conan.tools.cmake import cmake_layout

    class Pkg(ConanFile):
        name = "foo"
        version = "1.0"
        settings = "os", "arch", "build_type", "compiler"
        generators = "CMakeToolchain"
        exports_sources = "CMakeLists.txt", "main.cpp"

        def build(self):
            cmake = CMake(self)

            config_stdout, config_stderr = StringIO(), StringIO()
            cmake.configure(stdout=config_stdout, stderr=config_stderr)
            self.output.info(f"Configure stdout: '{config_stdout.getvalue()}'")
            self.output.info(f"Configure stderr: '{config_stderr.getvalue()}'")

            build_stdout, build_stderr = StringIO(), StringIO()
            cmake.build(stdout=build_stdout, stderr=build_stderr)
            self.output.info(f"Build stdout: '{build_stdout.getvalue()}'")
            self.output.info(f"Build stderr: '{build_stderr.getvalue()}'")

            test_stdout, test_stderr = StringIO(), StringIO()
            cmake.test(stdout=test_stdout, stderr=test_stderr)
            self.output.info(f"Test stdout: '{test_stdout.getvalue()}'")
            self.output.info(f"Test stderr: '{test_stderr.getvalue()}'")

        def package(self):
            cmake = CMake(self)
            install_stdout, install_stderr = StringIO(), StringIO()
            cmake.install(stdout=install_stdout, stderr=install_stderr)
            self.output.info(f"Install stdout: '{install_stdout.getvalue()}'")
            self.output.info(f"Install stderr: '{install_stderr.getvalue()}'")


    """)
    client.save({"conanfile.py": conanfile,
                 "CMakeLists.txt": 'project(foo)\nadd_executable(mylib main.cpp)\ninclude(CTest)',
                 "main.cpp": "int main() {return 0;}"})
    client.run("create .")

    # Ensure the output is not unexpectedly empty
    assert re.search("Configure stdout: '[^']", client.out)
    assert re.search("Configure stderr: '[^']", client.out)

    assert re.search("Build stdout: '[^']", client.out)
    assert re.search("Build stderr: ''", client.out)

    assert re.search("Test stdout: '[^']", client.out)

    # Empty on Windows
    if platform.system() == "Windows":
        assert re.search("Test stderr: ''", client.out)
    else:
        assert re.search("Test stderr: '[^']", client.out)

    if platform.system() == "Windows":
        assert re.search("Install stdout: ''", client.out)
    else:
        assert re.search("Install stdout: '[^']", client.out)
    assert re.search("Install stderr: ''", client.out)


@pytest.mark.tool("cmake")
@pytest.mark.skipif(platform.system() != "Windows", reason="neeed multi-config")
def test_cmake_toolchain_cxxflags_multi_config():
    c = TestClient()
    profile_release = textwrap.dedent(r"""
        include(default)
        [conf]
        tools.build:defines=["conan_test_answer=42", "conan_test_other=24"]
        tools.build:cxxflags=["/Zc:__cplusplus"]
        """)
    profile_debug = textwrap.dedent(r"""
        include(default)
        [settings]
        build_type=Debug
        [conf]
        tools.build:defines=["conan_test_answer=123"]
        tools.build:cxxflags=["/W4"]
        """)

    conanfile = textwrap.dedent(r'''
        from conan import ConanFile
        from conan.tools.cmake import CMake, CMakeToolchain, cmake_layout

        class Test(ConanFile):
            exports_sources = "CMakeLists.txt", "src/*"
            settings = "os", "compiler", "arch", "build_type"
            generators = "CMakeToolchain"

            def layout(self):
                cmake_layout(self)

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
        ''')

    main = textwrap.dedent(r"""
        #include <iostream>
        #include <stdio.h>

        #define STR(x)   #x
        #define SHOW_DEFINE(x) printf("DEFINE %s=%s!\n", #x, STR(x))

        int main() {
            SHOW_DEFINE(conan_test_answer);
            #ifdef conan_test_other
            SHOW_DEFINE(conan_test_other);
            #endif
            char a = 123L;  // to trigger warnings

            #if __cplusplus
            std::cout << "CPLUSPLUS: __cplusplus" << __cplusplus<< "\n";
            #endif
        }
        """)

    cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.15)
        project(Test CXX)
        add_executable(example src/main.cpp)
        """)

    c.save({"conanfile.py": conanfile,
            "profile_release": profile_release,
            "profile_debug": profile_debug,
            "src/main.cpp": main,
            "CMakeLists.txt": cmakelists}, clean_first=True)
    c.run("install . -pr=./profile_release")
    c.run("install . -pr=./profile_debug")

    with c.chdir("build"):
        c.run_command("cmake .. -DCMAKE_TOOLCHAIN_FILE=generators/conan_toolchain.cmake")
        c.run_command("cmake --build . --config Release")
        assert "warning C4189" not in c.out
        c.run_command("cmake --build . --config Debug")
        assert "warning C4189" in c.out

    c.run_command(r"build\Release\example.exe")
    assert 'DEFINE conan_test_answer=42!' in c.out
    assert 'DEFINE conan_test_other=24!' in c.out
    assert "CPLUSPLUS: __cplusplus20" in c.out

    c.run_command(r"build\Debug\example.exe")
    assert 'DEFINE conan_test_answer=123' in c.out
    assert 'other=' not in c.out
    assert "CPLUSPLUS: __cplusplus19" in c.out


@pytest.mark.tool("ninja")
@pytest.mark.tool("cmake", "3.23")
def test_cmake_toolchain_ninja_multi_config():
    c = TestClient()
    profile_release = textwrap.dedent(r"""
        include(default)
        [conf]
        tools.cmake.cmaketoolchain:generator=Ninja Multi-Config
        tools.build:defines=["conan_test_answer=42", "conan_test_other=24"]
        """)
    profile_debug = textwrap.dedent(r"""
        include(default)
        [settings]
        build_type=Debug
        [conf]
        tools.cmake.cmaketoolchain:generator=Ninja Multi-Config
        tools.build:defines=["conan_test_answer=123"]
        """)
    profile_relwithdebinfo = textwrap.dedent(r"""
        include(default)
        [settings]
        build_type=RelWithDebInfo
        [conf]
        tools.cmake.cmaketoolchain:generator=Ninja Multi-Config
        tools.build:defines=["conan_test_answer=456", "conan_test_other=abc", 'conan_test_complex="1 2"']
        """)

    conanfile = textwrap.dedent(r'''
        from conan import ConanFile
        from conan.tools.cmake import CMake, CMakeToolchain, cmake_layout

        class Test(ConanFile):
            exports_sources = "CMakeLists.txt", "src/*"
            settings = "os", "compiler", "arch", "build_type"
            generators = "CMakeToolchain"

            def layout(self):
                cmake_layout(self)

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
        ''')

    main = textwrap.dedent(r"""
        #include <iostream>
        #include <stdio.h>

        #define STR(x)   #x
        #define SHOW_DEFINE(x) printf("DEFINE %s=%s!\n", #x, STR(x))

        int main() {
            SHOW_DEFINE(conan_test_answer);
            #ifdef conan_test_other
            SHOW_DEFINE(conan_test_other);
            #endif
            #ifdef conan_test_complex
            SHOW_DEFINE(conan_test_complex);
            #endif
        }
        """)

    cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.15)
        project(Test CXX)
        add_executable(example src/main.cpp)
        """)

    c.save({"conanfile.py": conanfile,
            "profile_release": profile_release,
            "profile_debug": profile_debug,
            "profile_relwithdebinfo": profile_relwithdebinfo,
            "src/main.cpp": main,
            "CMakeLists.txt": cmakelists})
    c.run("install . -pr=./profile_release")
    c.run("install . -pr=./profile_debug")
    c.run("install . -pr=./profile_relwithdebinfo")

    with c.chdir("build"):
        env = r".\generators\conanbuild.bat &&" if platform.system() == "Windows" else ""
        c.run_command(f"{env} cmake .. -G \"Ninja Multi-Config\" "
                      "-DCMAKE_TOOLCHAIN_FILE=generators/conan_toolchain.cmake")
        c.run_command(f"{env} cmake --build . --config Release")
        c.run_command(f"{env} cmake --build . --config Debug")
        c.run_command(f"{env} cmake --build . --config RelWithDebInfo")

    c.run_command(os.sep.join([".", "build", "Release", "example"]))
    assert 'DEFINE conan_test_answer=42!' in c.out
    assert 'DEFINE conan_test_other=24!' in c.out
    assert 'complex=' not in c.out

    c.run_command(os.sep.join([".", "build", "Debug", "example"]))
    assert 'DEFINE conan_test_answer=123' in c.out
    assert 'other=' not in c.out
    assert 'complex=' not in c.out

    c.run_command(os.sep.join([".", "build", "RelWithDebInfo", "example"]))
    assert 'DEFINE conan_test_answer=456!' in c.out
    assert 'DEFINE conan_test_other=abc!' in c.out
    assert 'DEFINE conan_test_complex="1 2"!' in c.out


@pytest.mark.tool("cmake")
@pytest.mark.skipif(platform.system() != "Darwin", reason="Only needs to run once, no need for extra platforms")
def test_cxx_version_not_overriden_if_hardcoded():
    """Any C++ standard set in the CMakeLists.txt will have priority even if the
    compiler.cppstd is set in the profile"""
    tc = TestClient()
    tc.run("new cmake_exe -dname=foo -dversion=1.0")

    cml_contents = tc.load("CMakeLists.txt")
    cml_contents = cml_contents.replace("project(foo CXX)\n", "project(foo CXX)\nset(CMAKE_CXX_STANDARD 17)\n")
    tc.save({"CMakeLists.txt": cml_contents})

    # The compiler.cppstd will not override the CXX_STANDARD variable
    tc.run("create . -s=compiler.cppstd=11")
    assert "Conan toolchain: C++ Standard 11 with extensions OFF" in tc.out
    assert "Warning: Standard CMAKE_CXX_STANDARD value defined in conan_toolchain.cmake to 11 has been modified to 17" in tc.out

    # Even though Conan warns, the compiled code is C++17
    assert "foo/1.0: __cplusplus2017" in tc.out

    tc.run("create . -s=compiler.cppstd=17")
    assert "Conan toolchain: C++ Standard 17 with extensions OFF" in tc.out
    assert "Warning: Standard CMAKE_CXX_STANDARD value defined in conan_toolchain.cmake to 17 has been modified to 17" not in tc.out


@pytest.mark.tool("cmake", "3.23")  # Android complains if <3.19
@pytest.mark.tool("android_ndk")
@pytest.mark.skipif(platform.system() != "Darwin", reason="NDK only installed on MAC")
def test_cmake_toolchain_crossbuild_set_cmake_compiler():
    # To reproduce https://github.com/conan-io/conan/issues/16960
    # the issue happens when you set CMAKE_CXX_COMPILER and CMAKE_C_COMPILER as cache variables
    # and then cross-build
    c = TestClient()

    ndk_path = tools_locations["android_ndk"]["system"]["path"][platform.system()]
    bin_path = ndk_path + (
        "/toolchains/llvm/prebuilt/darwin-x86_64/bin" if platform.machine() == "x86_64" else
        "/toolchains/llvm/prebuilt/darwin-arm64/bin"
    )

    android = textwrap.dedent(f"""
       [settings]
       os=Android
       os.api_level=23
       arch=x86_64
       compiler=clang
       compiler.version=12
       compiler.libcxx=c++_shared
       build_type=Release
       [conf]
       tools.android:ndk_path={ndk_path}
       tools.build:compiler_executables = {{"c": "{bin_path}/x86_64-linux-android23-clang", "cpp": "{bin_path}/x86_64-linux-android23-clang++"}}
       """)

    conanfile = textwrap.dedent(f"""
        from conan import ConanFile
        from conan.tools.cmake import CMake, CMakeToolchain, cmake_layout
        from conan.tools.build import check_min_cppstd
        class SDKConan(ConanFile):
            name = "sdk"
            version = "1.0"
            generators = "CMakeDeps", "VirtualBuildEnv"
            settings = "os", "compiler", "arch", "build_type"
            exports_sources = "CMakeLists.txt"
            def layout(self):
                cmake_layout(self)
            def generate(self):
                tc = CMakeToolchain(self)
                tc.cache_variables["SDK_VERSION"] = str(self.version)
                tc.generate()
            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
    """)

    cmake = textwrap.dedent(f"""
        cmake_minimum_required(VERSION 3.23)
        project(sdk VERSION ${{SDK_VERSION}}.0 LANGUAGES NONE)
        message("sdk: ${{SDK_VERSION}}.0")
    """)

    c.save({"android": android,
            "conanfile.py": conanfile,
            "CMakeLists.txt": cmake,})
    # first run works ok
    c.run('build . --profile:host=android')
    assert 'sdk: 1.0.0' in c.out
    # in second run CMake says that you have changed variables that require your cache to be deleted.
    # and deletes the cache and fails
    c.run('build . --profile:host=android')
    assert 'VERSION ".0" format invalid.' not in c.out
    assert 'sdk: 1.0.0' in c.out


@pytest.mark.tool("cmake")
def test_cmake_toolchain_language_c():
    client = TestClient()

    conanfile = textwrap.dedent(r"""
        from conan import ConanFile
        from conan.tools.cmake import CMake, CMakeToolchain, cmake_layout

        class Pkg(ConanFile):
            settings = "os", "compiler", "arch", "build_type"
            generators = "CMakeToolchain"
            languages = "C"

            def layout(self):
                cmake_layout(self)

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
        """)

    cmakelists = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.15)
        project(pkg NONE)
        """)

    client.save(
        {
            "conanfile.py": conanfile,
            "CMakeLists.txt": cmakelists,
        },
        clean_first=True,
    )

    if platform.system() == "Windows":
        # compiler.version=191 is already the default now
        client.run("build . -s compiler.cstd=11 -s compiler.version=191", assert_error=True)
        assert "The provided compiler.cstd=11 is not supported by msvc 191. Supported values are: []" \
               in client.out
    else:
        client.run("build . -s compiler.cppstd=11")
        # It doesn't fail


@pytest.mark.skipif(platform.system() != "Windows", reason="Only for windows")
@pytest.mark.tool("cmake")
def test_cmaketoolchain_check_function_exists():
    """
    covers the problematic check_function_exists() CMake DISCOURAGED check, that can fail
    for Release configuration, due to CMake issues. This was broken when we tried to
    add the automatic CMAKE_TRY_COMPILE_CONFIGURATION in release cases, as CheckFunctionExists.c
    is optimized away or fails to compile in Release, but works in Debug.
    This test is here to capture this legacy and problematic use case NOT to recommended this
    as a good practice, rather then opposite, DONT DO IT
    """

    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake, CMakeToolchain
        class Conan(ConanFile):
            settings = "os", "compiler", "arch", "build_type"
            def generate(self):
                tc = CMakeToolchain(self)
                # This gets back to older default behavior
                # tc.blocks["try_compile"].values["config"] = None
                tc.generate()
            def build(self):
                CMake(self).configure()
        """)
    consumer = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.15)
        project(MyHello C)
        include(CheckFunctionExists)
        check_function_exists(pow HAVE_POW)
        if(NOT HAVE_POW)
            message(FATAL_ERROR "Not math!!!!!")
        endif()
       """)
    c.save({"conanfile.py": conanfile,
            "CMakeLists.txt": consumer})

    c.run("build . ")  # Release breaks the check
    assert "Looking for pow - found" in c.out
    # And it no longer crashes


@pytest.mark.skipif(platform.system() != "Linux", reason="Linker scripts in Linux only")
def test_cmake_linker_scripts():
    conanfile = textwrap.dedent("""
       from conan import ConanFile
       from conan.tools.cmake import CMake, CMakeToolchain, cmake_layout

       class Test(ConanFile):
           exports_sources = "CMakeLists.txt", "src/*"
           settings = "os", "compiler", "arch", "build_type"
           generators = "CMakeToolchain"

           def layout(self):
               cmake_layout(self)

           def build(self):
               cmake = CMake(self)
               cmake.configure()
               cmake.build()
       """)

    main = "int main() {}"

    cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.15)
        project(Test CXX)
        add_executable(example main.cpp)
        """)
    profile = textwrap.dedent("""
       include(default)
       [conf]
       tools.build:linker_scripts=['{{profile_dir}}/mylinkscript.ld']
       """)

    c = TestClient()
    c.save({"conanfile.py": conanfile,
            "CMakeLists.txt": cmakelists,
            "main.cpp": main,
            "mylinkscript.ld": "",
            "profile": profile})
    c.run('build . -pr=profile', assert_error=True)
    # This error means the linker script was fonud and loaded, it failed because empty
    assert "PHDR segment not covered by LOAD segment" in c.out


def test_cmake_toolchain_verbosity_propagation():
    t = TestClient(light=True)
    t.save({"conanfile.py": GenConanfile("mylib", "1.0")})
    t.run("create .")
    t.run("install --requires=mylib/1.0 -g CMakeToolchain")
    toolchain = t.load("conan_toolchain.cmake")
    assert 'set(CMAKE_VERBOSE_MAKEFILE' not in toolchain
    assert 'set(CMAKE_MESSAGE_LOG_LEVEL' not in toolchain

    t.run("install --requires=mylib/1.0 -g CMakeToolchain -c tools.build:verbosity=verbose "
          "-c tools.compilation:verbosity=verbose")
    toolchain = t.load("conan_toolchain.cmake")
    assert 'set(CMAKE_VERBOSE_MAKEFILE "ON"' in toolchain
    assert 'set(CMAKE_MESSAGE_LOG_LEVEL "VERBOSE"' in toolchain

    t.run("install --requires=mylib/1.0 -g CMakeToolchain -c tools.build:verbosity=quiet "
          "-c tools.compilation:verbosity=quiet")
    toolchain = t.load("conan_toolchain.cmake")
    assert 'set(CMAKE_MESSAGE_LOG_LEVEL "ERROR"' in toolchain

    # Extra variables have preference
    t.run("install --requires=mylib/1.0 -g CMakeToolchain -c tools.compilation:verbosity=quiet "
          "-c tools.cmake.cmaketoolchain:extra_variables=\"{'CMAKE_MESSAGE_LOG_LEVEL': 'WARNING'}\"")
    toolchain = t.load("conan_toolchain.cmake")
    # assert 'set(CMAKE_VERBOSE_MAKEFILE "OFF"' in toolchain
    assert 'set(CMAKE_MESSAGE_LOG_LEVEL "WARNING"' in toolchain



import platform
import textwrap

import pytest

from conan.test.utils.tools import TestClient


@pytest.mark.skipif(platform.system() != "Windows", reason="Fails on windows")
def test_crossbuild_windows_incomplete():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "os", "compiler", "build_type", "arch"
            def build(self):
                # using python --version as it does not depend on shell
                self.run("python --version")
            """)
    build_profile = textwrap.dedent("""
        [settings]
        arch=x86_64
    """)
    client.save({"conanfile.py": conanfile, "build_profile": build_profile})
    client.run("create . -pr:b=build_profile", assert_error=True)
    assert "The 'build' profile must have a 'os' declared" in client.out

    build_profile = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
        """)
    client.save({"conanfile.py": conanfile, "build_profile": build_profile})
    client.run("create . -pr:b=build_profile")
    assert "Python" in client.out


def test_quoted_vars():
    # https://github.com/conan-io/conan/issues/18761
    c = TestClient()

    dep = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "dep"
            version = "0.1"
            settings = "os"
            def package_info(self):
                v = 'MyVar:"content!! | other!!"'
                v = v.replace("|", "^|") if self.settings.os == "Windows" else v  # escape
                self.buildenv_info.define("MyCustomEscapedVar", v)
        """)

    conanfile = textwrap.dedent("""
        import os, platform
        from conan import ConanFile

        class ConanFileToolsTest(ConanFile):
            settings = "os"
            requires = "dep/0.1"
            def build(self):
                if platform.system() == "Windows":
                    self.run("set MyCustomEscapedVar")
                else:
                    self.run("printenv MyCustomEscapedVar")
                self.run("echo Hello!!!!")
        """)

    c.save({"dep/conanfile.py": dep,
            "consumer/conanfile.py": conanfile})

    c.run("create dep")
    c.run("build consumer")
    assert 'MyVar:"content!! | other!!"' in c.out
    assert "Hello!!!!" in c.out



import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


@pytest.fixture
def client():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake

        class Pkg(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            generators = "CMakeToolchain"

            def run(self, cmd, env=None, **kwargs):  # INTERCEPTOR of running
                self.output.info("RECIPE-RUN: {}".format(cmd))

            def build(self):
                cmake = CMake(self)
                cmake.build()
        """)
    client.save({"conanfile.py": conanfile})
    return client


def test_cmake_no_config(client):
    profile = textwrap.dedent("""\
        [settings]
        os=Windows
        arch=x86_64
        compiler=msvc
        compiler.version=191
        compiler.runtime=dynamic
        build_type=Release
        """)
    client.save({"myprofile": profile})
    client.run("create . --name=pkg --version=0.1 -pr=myprofile")
    assert "-verbosity" not in client.out


def test_cmake_config(client):
    profile = textwrap.dedent("""\
        [settings]
        os=Windows
        arch=x86_64
        compiler=msvc
        compiler.version=191
        compiler.runtime=dynamic
        build_type=Release
        [conf]
        tools.build:verbosity=quiet
        """)
    client.save({"myprofile": profile})
    client.run("create . --name=pkg --version=0.1 -pr=myprofile")
    assert "-verbosity:Quiet" in client.out
    client.run("create . --name=pkg --version=0.1 -pr=myprofile -c=tools.env.virtualenv:powershell=powershell.exe")
    assert "/verbosity:Quiet" in client.out


def test_cmake_config_error(client):
    profile = textwrap.dedent("""\
        [settings]
        os=Windows
        arch=x86_64
        compiler=msvc
        compiler.version=191
        compiler.runtime=dynamic
        build_type=Release
        [conf]
        tools.build:verbosity=non-existing
        """)
    client.save({"myprofile": profile})
    client.run("create . --name=pkg --version=0.1 -pr=myprofile", assert_error=True)
    assert "Unknown value 'non-existing' for 'tools.build:verbosity'" in client.out


def test_cmake_config_package(client):
    profile = textwrap.dedent("""\
        [settings]
        os=Windows
        arch=x86_64
        compiler=msvc
        compiler.version=191
        compiler.runtime=dynamic
        build_type=Release
        [conf]
        dep*:tools.build:verbosity=quiet
        """)
    client.save({"myprofile": profile})
    client.run("create . --name=pkg --version=0.1 -pr=myprofile")
    assert "-verbosity" not in client.out
    client.run("create . --name=dep --version=0.1 -pr=myprofile")
    assert "-verbosity:Quiet" in client.out


def test_cmake_config_package_not_scoped(client):
    profile = textwrap.dedent("""\
        [settings]
        os=Windows
        arch=x86_64
        compiler=msvc
        compiler.version=191
        compiler.runtime=dynamic
        build_type=Release
        [conf]
        tools.build:verbosity=quiet
        """)
    client.save({"myprofile": profile})
    client.run("create . --name=pkg --version=0.1 -pr=myprofile")
    assert "-verbosity:Quiet" in client.out
    client.run("create . --name=dep --version=0.1 -pr=myprofile")
    assert "-verbosity:Quiet" in client.out


def test_config_profile_forbidden(client):
    profile = textwrap.dedent("""\
        [conf]
        cache:verbosity=Minimal
        """)
    client.save({"myprofile": profile})
    client.run("install . --name=pkg --version=0.1 -pr=myprofile", assert_error=True)
    assert ("ERROR: Error reading 'myprofile' profile: [conf] "
            "'cache:verbosity' not allowed in profiles" in client.out)


def test_msbuild_config():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.microsoft import MSBuild

        class Pkg(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            def build(self):
                ms = MSBuild(self)
                self.output.info(ms.command("Project.sln"))
        """)
    client.save({"conanfile.py": conanfile})
    profile = textwrap.dedent("""\
        [settings]
        os=Windows
        arch=x86_64
        compiler=msvc
        compiler.version=191
        compiler.runtime=dynamic
        build_type=Release
        [conf]
        tools.build:verbosity=quiet
        """)
    client.save({"myprofile": profile})
    client.run("create . --name=pkg --version=0.1 -pr=myprofile")
    assert "-verbosity:Quiet" in client.out


def test_msbuild_compile_options():
    # This works in all platforms because MSBuildToolchain works even in Linux, it will
    # just skip generating the conanvcvars.bat
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            generators = "MSBuildToolchain"
        """)
    client.save({"conanfile.py": conanfile})

    profile = textwrap.dedent("""\
        [settings]
        os=Windows
        arch=x86_64
        compiler=msvc
        compiler.version=191
        compiler.runtime=dynamic
        build_type=Release
        [conf]
        tools.microsoft.msbuildtoolchain:compile_options={"ExceptionHandling": "Async"}
        """)
    client.save({"myprofile": profile})
    client.run("install . -pr=myprofile")
    msbuild_tool = client.load("conantoolchain_release_x64.props")
    assert "<ExceptionHandling>Async</ExceptionHandling>" in msbuild_tool


def test_conf_package_patterns():
    client = TestClient()
    conanfile = GenConanfile()
    generate = """
    def generate(self):
        value = self.conf.get("user.build:myconfig")
        self.output.warning("{} Config:{}".format(self.ref.name, value))
"""
    client.save({"dep/conanfile.py": str(conanfile) + generate,
                 "pkg/conanfile.py": str(conanfile.with_requirement("dep/0.1", visible=False)) + generate,
                 "consumer/conanfile.py": str(conanfile.with_requires("pkg/0.1")
                .with_settings("os", "build_type")) + generate})

    client.run("export dep --name=dep --version=0.1")
    client.run("export pkg --name=pkg --version=0.1")

    # This pattern applies to no package
    profile = """
    [settings]
    os=Windows
    [conf]
    invented/*:user.build:myconfig=Foo
    """
    client.save({"profile": profile})
    client.run("install consumer --build=* --profile profile")
    assert "WARN: dep Config:None" in client.out
    assert "WARN: pkg Config:None" in client.out
    assert "WARN: None Config:None" in client.out

    # This patterns applies to dep
    profile = """
    [settings]
    os=Windows
    [conf]
    dep/*:user.build:myconfig=Foo
    """
    client.save({"profile": profile})
    client.run("install consumer --build='*' --profile profile")
    assert "WARN: dep Config:Foo" in client.out
    assert "WARN: pkg Config:None" in client.out
    assert "WARN: None Config:None" in client.out

    profile = """
    [settings]
    os=Windows
    [conf]
    dep/0.1:user.build:myconfig=Foo
    """
    client.save({"profile": profile})
    client.run("install consumer --build='*' --profile profile")
    assert "WARN: dep Config:Foo" in client.out
    assert "WARN: pkg Config:None" in client.out
    assert "WARN: None Config:None" in client.out

    # The global pattern applies to all
    profile = """
    [settings]
    os=Windows
    [conf]
    dep/*:user.build:myconfig=Foo
    pkg/*:user.build:myconfig=Foo2
    user.build:myconfig=Var
    """
    client.save({"profile": profile})
    client.run("install consumer --build='*' --profile profile")
    assert "WARN: dep Config:Var" in client.out
    assert "WARN: pkg Config:Var" in client.out
    assert "WARN: None Config:Var" in client.out

    # "&" pattern for the consumer
    profile = """
    [settings]
    os=Windows
    [conf]
    dep/*:user.build:myconfig=Foo
    pkg/*:user.build:myconfig=Foo2
    &:user.build:myconfig=Var
    """

    client.save({"profile": profile})
    client.run("install consumer --build='*' --profile profile")
    assert "WARN: dep Config:Foo" in client.out
    assert "WARN: pkg Config:Foo2" in client.out
    assert "WARN: None Config:Var" in client.out


def test_config_package_append(client):
    profile1 = textwrap.dedent("""\
        [conf]
        user.myteam:myconf=["a", "b", "c"]
        """)
    profile2 = textwrap.dedent("""\
        include(profile1)
        [conf]
        mypkg*:user.myteam:myconf+=["d"]
        mydep*:user.myteam:myconf=+["e"]
        """)
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            def generate(self):
                self.output.info(f"MYCONF: {self.conf.get('user.myteam:myconf')}")
            def build(self):
                self.output.info(f"MYCONFBUILD: {self.conf.get('user.myteam:myconf')}")
            """)
    client.save({"profile1": profile1,
                 "profile2": profile2,
                 "conanfile.py": conanfile})
    client.run("install . --name=mypkg --version=0.1 -pr=profile2")
    assert "conanfile.py (mypkg/0.1): MYCONF: ['a', 'b', 'c', 'd']" in client.out
    client.run("install . --name=mydep --version=0.1 -pr=profile2")
    assert "conanfile.py (mydep/0.1): MYCONF: ['e', 'a', 'b', 'c']" in client.out

    client.run("create . --name=mypkg --version=0.1 -pr=profile2")
    assert "mypkg/0.1: MYCONFBUILD: ['a', 'b', 'c', 'd']" in client.out
    client.run("create . --name=mydep --version=0.1 -pr=profile2")
    assert "mydep/0.1: MYCONFBUILD: ['e', 'a', 'b', 'c']" in client.out


def test_conf_patterns_user_channel():
    # https://github.com/conan-io/conan/issues/14139
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):
            def configure(self):
                self.output.info(f"CONF: {self.conf.get('user.myteam:myconf')}")
                self.output.info(f"CONF2: {self.conf.get('user.myteam:myconf2')}")
        """)
    profile = textwrap.dedent("""\
        [conf]
        user.myteam:myconf=myvalue1
        user.myteam:myconf2=other1
        *@user/channel:user.myteam:myconf=myvalue2
        *@*/*:user.myteam:myconf2=other2
        """)
    client.save({"dep/conanfile.py": conanfile,
                 "app/conanfile.py": GenConanfile().with_requires("dep1/0.1",
                                                                  "dep2/0.1@user/channel"),
                 "profile": profile})

    client.run("create dep --name=dep1 --version=0.1")
    client.run("create dep --name=dep2 --version=0.1 --user=user --channel=channel")
    client.run("install app -pr=profile")
    assert "dep1/0.1: CONF: myvalue1" in client.out
    assert "dep2/0.1@user/channel: CONF: myvalue2" in client.out
    assert "dep1/0.1: CONF2: other1" in client.out
    assert "dep2/0.1@user/channel: CONF2: other2" in client.out



from requests import ConnectionError

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient, TestRequester


def test_upload_parallel_error():
    """Cause an error in the parallel transfer and see some message"""

    class FailOnReferencesUploader(TestRequester):
        fail_on = ["lib1", "lib3"]

        def put(self, *args, **kwargs):
            if any(ref in args[0] for ref in self.fail_on):
                raise ConnectionError("Connection fails with lib2 and lib4 references!")
            else:
                return super(FailOnReferencesUploader, self).put(*args, **kwargs)

    client = TestClient(requester_class=FailOnReferencesUploader, default_server_user=True)
    client.save_home({"global.conf": f"core.upload:parallel=2\ncore.upload:retry_wait=0"})
    client.save({"conanfile.py": GenConanfile()})
    client.run('remote login default admin -p password')
    for index in range(4):
        client.run('create . --name=lib{} --version=1.0 --user=user --channel=channel'.format(index))
    client.run('upload lib* -c -r default', assert_error=True)
    assert "Connection fails with lib2 and lib4 references!" in client.out
    assert "Execute upload again to retry upload the failed files" in client.out


def test_upload_parallel_success():
    """Upload 2 packages in parallel with success"""

    client = TestClient(default_server_user=True)
    client.save_home({"global.conf": f"core.upload:parallel=2"})
    client.save({"conanfile.py": GenConanfile()})
    client.run('create . --name=lib0 --version=1.0 --user=user --channel=channel')
    client.run('create . --name=lib1 --version=1.0 --user=user --channel=channel')
    client.run('remote login default admin -p password')
    client.run('upload lib* -c -r default')
    assert "Uploading recipe 'lib0/1.0@user/channel#4d670581ccb765839f2239cc8dff8fbd'" in client.out
    assert "Uploading recipe 'lib1/1.0@user/channel#4d670581ccb765839f2239cc8dff8fbd'" in client.out
    client.run('search lib0/1.0@user/channel -r default')
    assert "lib0/1.0@user/channel" in client.out
    client.run('search lib1/1.0@user/channel -r default')
    assert "lib1/1.0@user/channel" in client.out


def test_upload_parallel_fail_on_interaction():
    """Upload 2 packages in parallel and fail because non_interactive forced"""
    client = TestClient(default_server_user=True)
    client.save_home({"global.conf": f"core.upload:parallel=2\ncore:non_interactive=True"})
    client.save({"conanfile.py": GenConanfile()})
    num_references = 2
    for index in range(num_references):
        client.run('create . --name=lib{} --version=1.0 --user=user --channel=channel'.format(index))

    client.run('remote logout default')
    client.run('upload lib* -c -r default', assert_error=True)
    assert "ERROR: Conan interactive mode disabled. [Remote: default]" in client.out



import platform
import textwrap

import pytest

from conan.internal.model.pkg_type import PackageType
from conan.test.utils.tools import TestClient

new_value = "will_break_next"


@pytest.mark.parametrize("shared", [True, False])
@pytest.mark.tool("cmake", "3.27")
@pytest.mark.skipif(platform.system() != "Darwin", reason="Only OSX")
def test_osx_frameworks(shared):
    """
    Testing custom package frameworks + system frameworks + requirements
    """
    client = TestClient()
    client.run("new cmake_lib -d name=dep -d version=1.0")
    client.run(f"create . -tf='' -o '&:shared={shared}'")
    cmakelists = textwrap.dedent("""
    set(CMAKE_CXX_COMPILER_WORKS 1)
    set(CMAKE_CXX_ABI_COMPILED 1)
    cmake_minimum_required(VERSION 3.15)
    project(MyFramework CXX)

    find_package(dep CONFIG REQUIRED)

    add_library(MyFramework frame.cpp frame.h)

    set_target_properties(MyFramework PROPERTIES
      FRAMEWORK TRUE
      FRAMEWORK_VERSION C # Version "A" is macOS convention
      MACOSX_FRAMEWORK_IDENTIFIER MyFramework
      PUBLIC_HEADER frame.h
    )
    target_link_libraries(MyFramework PRIVATE dep::dep)

    if(BUILD_SHARED_LIBS)
        target_link_libraries(MyFramework PRIVATE "-framework CoreFoundation")
    endif()

    install(TARGETS MyFramework
      FRAMEWORK DESTINATION .)
    """)
    frame_cpp = textwrap.dedent("""
    #include "frame.h"
    #include "dep.h"
    #include <iostream>
    #include <CoreFoundation/CoreFoundation.h>

    void greet() {
        // CoreFoundation
        CFTypeRef keys[] = {CFSTR("key")};
        CFTypeRef values[] = {CFSTR("value")};
        CFDictionaryRef dict = CFDictionaryCreate(kCFAllocatorDefault, keys, values, sizeof(keys) / sizeof(keys[0]),
                        &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
        if (dict)
            CFRelease(dict);
        // MyFramework
        std::cout << "Hello from MyFramework!" << std::endl;

        // dep requirement
        dep();
    }
    """)
    frame_h = textwrap.dedent("""
    #pragma once
    #include <vector>
    #include <string>
    void greet();
    """)
    cpp_info_type = PackageType.SHARED if shared else PackageType.STATIC
    conanfile = textwrap.dedent(f"""
    import os
    from conan import ConanFile
    from conan.tools.cmake import CMake

    class MyFramework(ConanFile):
        name = "frame"
        version = "1.0"
        settings = "os", "arch", "compiler", "build_type"
        options = {{"shared": [True, False], "fPIC": [True, False]}}
        default_options = {{"shared": False, "fPIC": True}}
        exports_sources = "frame.cpp", "frame.h", "CMakeLists.txt"
        generators = "CMakeToolchain", "CMakeConfigDeps"
        requires = "dep/1.0"

        def config_options(self):
            if self.settings.os == "Windows":
                self.options.rm_safe("fPIC")

        def configure(self):
            if self.options.shared:
                self.options.rm_safe("fPIC")

        def build(self):
            cmake = CMake(self)
            cmake.configure()
            cmake.build()

        def package(self):
            cmake = CMake(self)
            cmake.install()

        def package_info(self):
            self.cpp_info.type = "{cpp_info_type}"
            self.cpp_info.package_framework = "MyFramework"
            self.cpp_info.location = os.path.join(self.package_folder, "MyFramework.framework", "MyFramework")
            # Using also a system framework
            self.cpp_info.frameworks = ["CoreFoundation"]
    """)
    test_main_cpp = textwrap.dedent("""
    #include <MyFramework/frame.h>
    int main() {
        greet();
    }
    """)
    test_conanfile = textwrap.dedent("""
    import os
    from conan import ConanFile
    from conan.tools.cmake import CMake, cmake_layout
    from conan.tools.build import can_run

    class LibTestConan(ConanFile):
        settings = "os", "compiler", "build_type", "arch"
        generators = "CMakeConfigDeps", "CMakeToolchain"

        def requirements(self):
            self.requires(self.tested_reference_str)

        def build(self):
            cmake = CMake(self)
            cmake.configure()
            cmake.build()

        def layout(self):
            cmake_layout(self)

        def test(self):
            if can_run(self):
                cmd = os.path.join(self.cpp.build.bindir, "example")
                self.run(cmd, env="conanrun")
    """)
    test_cmakelists = textwrap.dedent("""
    set(CMAKE_CXX_COMPILER_WORKS 1)
    set(CMAKE_CXX_ABI_COMPILED 1)
    cmake_minimum_required(VERSION 3.15)
    project(PackageTest CXX)
    find_package(frame CONFIG REQUIRED)
    add_executable(example main.cpp)
    target_link_libraries(example frame::frame)
    """)
    client.save({
        'test_package/main.cpp': test_main_cpp,
        'test_package/CMakeLists.txt': test_cmakelists,
        'test_package/conanfile.py': test_conanfile,
        'CMakeLists.txt': cmakelists,
        'frame.cpp': frame_cpp,
        'frame.h': frame_h,
        'conanfile.py': conanfile
    }, clean_first=True)
    client.run(f"create . -c tools.cmake.cmakedeps:new={new_value} -o '*:shared={shared}'")
    assert "Hello from MyFramework!" in client.out
    assert "dep/1.0: Hello World" in client.out



import os
import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient, NO_SETTINGS_PACKAGE_ID
from conan.internal.util.files import save, load


class TestMetadataCommands:

    @pytest.fixture
    def create_conan_pkg(self):
        client = TestClient(default_server_user=True, light=True)
        client.save({"conanfile.py": GenConanfile("pkg", "0.1")})
        client.run("create .")
        pid = client.created_package_id("pkg/0.1")
        return client, pid

    @staticmethod
    def save_metadata_file(client, pkg_ref, filename="somefile.log", content=None):
        client.run(f"cache path {pkg_ref} --folder=metadata")
        metadata_path = str(client.stdout).strip()
        myfile = os.path.join(metadata_path, "logs", filename)
        save(myfile, f"{content or str(pkg_ref)}!!!!")
        return metadata_path, myfile

    def test_upload(self, create_conan_pkg):
        c, pid = create_conan_pkg

        # Add some metadata
        self.save_metadata_file(c, "pkg/0.1", "mylogs.txt")
        self.save_metadata_file(c, f"pkg/0.1:{pid}", "mybuildlogs.txt")
        # Now upload everything
        c.run("upload * -c -r=default")
        assert "pkg/0.1: Recipe metadata: 1 files" in c.out
        assert "pkg/0.1:da39a3ee5e6b4b0d3255bfef95601890afd80709: Package metadata: 1 files" in c.out

        # Add new files to the metadata
        self.save_metadata_file(c, "pkg/0.1", "mylogs2.txt")
        self.save_metadata_file(c, f"pkg/0.1:{pid}", "mybuildlogs2.txt")
        # Upload the metadata, even if the revisions exist in the server
        # adding the new metadata logs files
        c.run("upload * -c -r=default --metadata=*")
        assert "pkg/0.1: Recipe metadata: 2 files" in c.out
        assert "pkg/0.1:da39a3ee5e6b4b0d3255bfef95601890afd80709: Package metadata: 2 files" in c.out

        c.run("remove * -c")
        c.run("install --requires=pkg/0.1")  # wont install metadata by default
        c.run("cache path pkg/0.1 --folder=metadata")
        metadata_path = str(c.stdout).strip()
        assert os.listdir(metadata_path) == []
        c.run(f"cache path pkg/0.1:{pid} --folder=metadata")
        metadata_path = str(c.stdout).strip()
        assert os.listdir(metadata_path) == []

        # Forcing the download of the metadata of cache-existing things with the "download" command
        c.run("download pkg/0.1 -r=default --metadata=*")
        c.run(f"cache path pkg/0.1 --folder=metadata")
        metadata_path = str(c.stdout).strip()
        c.run(f"cache path pkg/0.1:{pid} --folder=metadata")
        pkg_metadata_path = str(c.stdout).strip()
        for f in "logs/mylogs.txt", "logs/mylogs2.txt":
            assert os.path.isfile(os.path.join(metadata_path, f))
        for f in "logs/mybuildlogs.txt", "logs/mybuildlogs2.txt":
            assert os.path.isfile(os.path.join(pkg_metadata_path, f))

    def test_update_contents(self):
        c = TestClient(default_server_user=True, light=True)
        c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
        c.run("export .")

        # Add some metadata
        _, myfile = self.save_metadata_file(c, "pkg/0.1", "mylogs.txt")

        # Now upload everything
        c.run("upload * -c -r=default")
        assert "pkg/0.1: Recipe metadata: 1 files" in c.out

        # Update the metadata
        save(myfile, "mylogs2!!!!")
        # Upload the metadata, even if the revisions exist in the server
        # adding the new metadata logs files
        c.run("upload * -c -r=default --metadata=*")
        assert "pkg/0.1: Recipe metadata: 1 files" in c.out

        c.run("remove * -c")
        c.run("download pkg/0.1 -r=default --metadata=*")
        c.run("cache path pkg/0.1 --folder=metadata")
        metadata_path = str(c.stdout).strip()

        content = load(os.path.join(metadata_path, "logs", "mylogs.txt"))
        assert "mylogs2!!!!" in content

    def test_append_contents(self):
        # I can add extra files to metadata without necessarily downloading it first
        c = TestClient(default_server_user=True, light=True)
        c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
        c.run("create .")
        pkgid = "da39a3ee5e6b4b0d3255bfef95601890afd80709"
        # Add some metadata
        self.save_metadata_file(c, "pkg/0.1", "mylogs.txt", content="mylogs")
        self.save_metadata_file(c, f"pkg/0.1:{pkgid}", "mybin.txt", content="mybin")

        # Now upload everything
        c.run("upload * -c -r=default")
        assert "pkg/0.1: Recipe metadata: 1 files" in c.out

        c2 = TestClient(servers=c.servers, light=True, inputs=["admin", "password"])
        c2.run("install --requires=pkg/0.1")
        self.save_metadata_file(c2, "pkg/0.1", "mylogs2.txt", content="mylogs2")
        self.save_metadata_file(c2, f"pkg/0.1:{pkgid}", "mybin2.txt", content="mybin2")
        c2.run("upload * -c -r=default --metadata=*")
        assert "pkg/0.1: Recipe metadata: 1 files" in c2.out

        c.run("remove * -c")
        c.run("download pkg/0.1 -r=default --metadata=*")

        c.run("cache path pkg/0.1 --folder=metadata")
        metadata_path = str(c.stdout).strip()
        assert "mylogs!!!!" in load(os.path.join(metadata_path, "logs", "mylogs.txt"))
        assert "mylogs2!!!!" in load(os.path.join(metadata_path, "logs", "mylogs2.txt"))

        c.run(f"cache path pkg/0.1:{pkgid} --folder=metadata")
        metadata_path = str(c.stdout).strip()
        assert "mybin!!!!" in load(os.path.join(metadata_path, "logs", "mybin.txt"))
        assert "mybin2!!!!" in load(os.path.join(metadata_path, "logs", "mybin2.txt"))

    def test_overwrite_server_contents(self):
        # I can overwrite server contents
        c = TestClient(default_server_user=True, light=True)
        c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
        c.run("create .")
        pkgid = "da39a3ee5e6b4b0d3255bfef95601890afd80709"
        # Add some metadata
        self.save_metadata_file(c, "pkg/0.1", "mylogs.txt", content="mylogs")
        self.save_metadata_file(c, f"pkg/0.1:{pkgid}", "mybin.txt", content="mybin")

        # Now upload everything
        c.run("upload * -c -r=default")
        assert "pkg/0.1: Recipe metadata: 1 files" in c.out

        c2 = TestClient(servers=c.servers, light=True, inputs=["admin", "password"])
        c2.run("install --requires=pkg/0.1")
        self.save_metadata_file(c2, "pkg/0.1", "mylogs.txt", content="mylogs2")
        self.save_metadata_file(c2, f"pkg/0.1:{pkgid}", "mybin.txt", content="mybin2")
        c2.run("upload * -c -r=default --metadata=*")
        assert "pkg/0.1: Recipe metadata: 1 files" in c2.out

        c.run("remove * -c")
        c.run("download pkg/0.1 -r=default --metadata=*")

        c.run("cache path pkg/0.1 --folder=metadata")
        metadata_path = str(c.stdout).strip()
        assert "mylogs2!!!!" in load(os.path.join(metadata_path, "logs", "mylogs.txt"))

        c.run(f"cache path pkg/0.1:{pkgid} --folder=metadata")
        metadata_path = str(c.stdout).strip()
        assert "mybin2!!!!" in load(os.path.join(metadata_path, "logs", "mybin.txt"))

    def test_folder_exist(self, create_conan_pkg):
        """ so we can cp -R to the metadata folder, having to create the folder in the cache
        is weird
        """
        c, _ = create_conan_pkg
        c.run("cache path pkg/0.1 --folder=metadata")
        metadata_path = str(c.stdout).strip()
        assert os.path.isdir(metadata_path)
        c.run(f"cache path pkg/0.1:{NO_SETTINGS_PACKAGE_ID} --folder=metadata")
        pkg_metadata_path = str(c.stdout).strip()
        assert os.path.isdir(pkg_metadata_path)

    def test_direct_download_redownload(self, create_conan_pkg):
        """ When we directly download things, without "conan install" first, it is also able
        to fetch the requested metadata

        Also, re-downloading same thing shouldn't fail
        """
        c, pid = create_conan_pkg

        # Add some metadata
        metadata_path, _ = self.save_metadata_file(c, "pkg/0.1", "mylogs.txt")
        self.save_metadata_file(c, f"pkg/0.1:{pid}", "mybuildlogs.txt")

        # Now upload everything
        c.run("upload * -c -r=default")
        assert "pkg/0.1: Recipe metadata: 1 files" in c.out
        assert "pkg/0.1:da39a3ee5e6b4b0d3255bfef95601890afd80709: Package metadata: 1 files" in c.out

        c.run("remove * -c")

        # Forcing the download of the metadata of cache-existing things with the "download" command
        c.run("download pkg/0.1 -r=default --metadata=*")
        assert os.path.isfile(os.path.join(metadata_path, "logs", "mylogs.txt"))
        c.run(f"cache path pkg/0.1:{pid} --folder=metadata")
        pkg_metadata_path = str(c.stdout).strip()
        assert os.path.isfile(os.path.join(pkg_metadata_path, "logs", "mybuildlogs.txt"))

        # Re-download shouldn't fail
        c.run("download pkg/0.1 -r=default --metadata=*")
        assert os.path.isfile(os.path.join(metadata_path, "logs", "mylogs.txt"))
        c.run(f"cache path pkg/0.1:{pid} --folder=metadata")
        pkg_metadata_path = str(c.stdout).strip()
        assert os.path.isfile(os.path.join(pkg_metadata_path, "logs", "mybuildlogs.txt"))

    def test_no_download_cached(self, create_conan_pkg):
        """ as the metadata can change, no checksum, no revision, cannot be cached
        """
        c, pid = create_conan_pkg

        # Add some metadata
        _, myrecipefile = self.save_metadata_file(c, "pkg/0.1", "mylogs.txt")
        _, mypkgfile = self.save_metadata_file(c, f"pkg/0.1:{pid}", "mybuildlogs.txt")

        # Now upload everything
        c.run("upload * -c -r=default")
        assert "pkg/0.1: Recipe metadata: 1 files" in c.out
        assert "pkg/0.1:da39a3ee5e6b4b0d3255bfef95601890afd80709: Package metadata: 1 files" in c.out

        c2 = TestClient(servers=c.servers, light=True)
        tmp_folder = temp_folder()
        # MOST important part: activate cache
        c2.save_home({"global.conf": f"core.download:download_cache={tmp_folder}\n"})

        # download package and metadata
        c2.run("download pkg/0.1 -r=default --metadata=*")
        c2.run("cache path pkg/0.1 --folder=metadata")
        c2_metadata_path = str(c2.stdout).strip()
        mylogs = load(os.path.join(c2_metadata_path, "logs", "mylogs.txt"))
        assert "pkg/0.1!!!!" in mylogs
        c2.run(f"cache path pkg/0.1:{pid} --folder=metadata")
        c2_pkg_metadata_path = str(c2.stdout).strip()
        mybuildlogs = load(os.path.join(c2_pkg_metadata_path, "logs", "mybuildlogs.txt"))
        assert f"pkg/0.1:{pid}!!!!" in mybuildlogs

        # Now the other client will update the metadata
        save(myrecipefile, "mylogs2!!!!")
        save(mypkgfile, "mybuildlogs2!!!!")
        c.run("upload * -c -r=default --metadata=*")
        assert "pkg/0.1: Recipe metadata: 1 files" in c.out
        assert "pkg/0.1:da39a3ee5e6b4b0d3255bfef95601890afd80709: Package metadata: 1 files" in c.out

        # re-download of metadata in c2
        c2.run("remove * -c")  # to make sure the download cache works
        c2.run("download pkg/0.1 -r=default --metadata=*")
        mylogs = load(os.path.join(c2_metadata_path, "logs", "mylogs.txt"))
        assert "mylogs2!!!!" in mylogs
        mybuildlogs = load(os.path.join(c2_pkg_metadata_path, "logs", "mybuildlogs.txt"))
        assert "mybuildlogs2!!!!" in mybuildlogs

    def test_upload_ignored_metadata(self, create_conan_pkg):
        """
        Upload command should ignore metadata files when passing --metadata=""
        """
        client, pid = create_conan_pkg

        self.save_metadata_file(client, "pkg/0.1")
        self.save_metadata_file(client, f"pkg/0.1:{pid}")

        client.run('upload * --confirm --remote=default --metadata=""')
        assert "Recipe metadata" not in client.out
        assert "Package metadata" not in client.out

    def test_upload_ignored_metadata_with_pattern(self, create_conan_pkg):
        """
        Upload command should fail when passing --metadata="" and a pattern
        """
        client = TestClient(default_server_user=True, light=True)
        client.save({"conanfile.py": GenConanfile("pkg", "0.1")})
        client.run("export .")

        client.run('upload * --confirm --remote=default --metadata="" --metadata="logs/*"',
                   assert_error=True)
        assert "ERROR: Empty string and patterns can not be mixed for metadata." in client.out



import sys
from unittest import mock

import pytest

from conan.api.output import ConanOutput, init_colorama
from conan.test.utils.mocks import RedirectedTestOutput
from conan.test.utils.tools import redirect_output


class TestConanOutput:

    @pytest.mark.parametrize("isatty, env", [(True, {"NO_COLOR": "1"}),
                                             (True, {"NO_COLOR": "0"})])
    def test_output_color_prevent_strip(self, isatty, env):
        with mock.patch("colorama.init") as init:
            with mock.patch("sys.stderr.isatty", return_value=isatty), \
                 mock.patch.dict("os.environ", env, clear=True):
                init_colorama(sys.stderr)
                out = ConanOutput()
                assert out.color is False
                init.assert_not_called()


@pytest.mark.parametrize("force", ["1", "0", "foo"])
def test_output_forced(force):
    env = {"CLICOLOR_FORCE": force}
    forced = force != "0"
    with mock.patch("colorama.init") as init:
        with mock.patch("sys.stderr.isatty", return_value=False), \
             mock.patch.dict("os.environ", env, clear=True):
            init_colorama(sys.stderr)
            out = ConanOutput()

            assert out.color is forced
            if not forced:
                init.assert_not_called()


def test_output_chainable():
    stderr = RedirectedTestOutput()
    with redirect_output(stderr):
        ConanOutput(scope="My package")\
            .title("My title")\
            .highlight("Worked")\
            .info("But there was more that needed to be said")
    assert "My package" in stderr.getvalue()
    assert "My title" in stderr.getvalue()
    assert "Worked" in stderr.getvalue()
    assert "But there was more that needed to be said" in stderr.getvalue()



import json
import os
import re
import textwrap

import pytest

from conan.api.model import RecipeReference
from conan.cli.exit_codes import ERROR_GENERAL
from conan.test.utils.tools import TestClient, NO_SETTINGS_PACKAGE_ID, GenConanfile
from conan.internal.util.files import load


def test_dependencies_order_matches_requires():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile()})
    client.run("create . --name=pkga --version=0.1 --user=user --channel=testing")
    client.save({"conanfile.py": GenConanfile()})
    client.run("create . --name=pkgb --version=0.1 --user=user --channel=testing")
    conanfile = textwrap.dedent("""
        [requires]
        pkgb/0.1@user/testing
        pkga/0.1@user/testing
        """)
    client.save({"conanfile.txt": conanfile}, clean_first=True)
    client.run("install . -g MSBuildDeps -s build_type=Release -s arch=x86")
    conandeps = client.load("conandeps.props")
    assert conandeps.find("pkgb") < conandeps.find("pkga")


def test_create():
    client = TestClient()
    conanfile = textwrap.dedent("""
    from conan import ConanFile
    class MyPkg(ConanFile):
        def source(self):
            assert(self.version=="0.1")
            assert(self.name=="pkg")
        def configure(self):
            assert(self.version=="0.1")
            assert(self.name=="pkg")
        def requirements(self):
            assert(self.version=="0.1")
            assert(self.name=="pkg")
        def build(self):
            assert(self.version=="0.1")
            assert(self.name=="pkg")
        def package(self):
            assert(self.version=="0.1")
            assert(self.name=="pkg")
        def package_info(self):
            assert(self.version=="0.1")
            assert(self.name=="pkg")
        def system_requirements(self):
            assert(self.version=="0.1")
            assert(self.name=="pkg")
            self.output.info("Running system requirements!!")
    """)
    client.save({"conanfile.py": conanfile})
    client.run("create --name=pkg --version=0.1 --user=lasote --channel=testing")
    assert "Profile host:\n[settings]" in client.out
    assert "pkg/0.1@lasote/testing: Generating the package" in client.out
    assert "Running system requirements!!" in client.out
    client.run('list -c *')
    assert "pkg/0.1@lasote/testing" in client.out

    # Create with only user will raise an error because of no name/version
    client.run("create conanfile.py --user=lasote --channel=testing", assert_error=True)
    assert "ERROR: conanfile didn't specify name" in client.out
    # Create with user but no channel should be valid
    client.run("create . --name=pkg --version=0.1 --user=lasote")
    assert "pkg/0.1@lasote:" in client.out


def test_error_create_name_version():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile().with_name("hello").with_version("1.2")})
    client.run("create . --name=hello --version=1.2 --user=lasote --channel=stable")
    client.run("create . --name=pkg", assert_error=True)
    assert "ERROR: Package recipe with name pkg!=hello" in client.out
    client.run("create . --version=1.1", assert_error=True)
    assert "ERROR: Package recipe with version 1.1!=1.2" in client.out


def test_create_user_channel():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile().with_name("pkg").with_version("0.1")})
    client.run("create . --user=lasote --channel=channel")
    assert "pkg/0.1@lasote/channel: Generating the package" in client.out
    client.run("list * -c")
    assert "pkg/0.1@lasote/channel" in client.out

    # test default without user and channel
    client.run("create . ")
    assert "pkg/0.1: Generating the package" in client.out


def test_create_in_subfolder():
    client = TestClient()
    client.save({"subfolder/conanfile.py": GenConanfile().with_name("pkg").with_version("0.1")})
    client.run("create subfolder --user=lasote --channel=channel")
    assert "pkg/0.1@lasote/channel: Generating the package" in client.out
    client.run("list * -c")
    assert "pkg/0.1@lasote/channel" in client.out


def test_create_in_subfolder_with_different_name():
    # Now with a different name
    client = TestClient()
    client.save({"subfolder/Custom.py": GenConanfile().with_name("pkg").with_version("0.1")})
    client.run("create subfolder/Custom.py --user=lasote --channel=channel")
    assert "pkg/0.1@lasote/channel: Generating the package" in client.out
    client.run("list * -c")
    assert "pkg/0.1@lasote/channel" in client.out


def test_create_test_package():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile().with_name("pkg").with_version("0.1"),
                 "test_package/conanfile.py":
                     GenConanfile().with_test('self.output.info("TESTING!!!")')})
    client.run("create . --user=lasote --channel=testing")
    assert "pkg/0.1@lasote/testing: Generating the package" in client.out
    assert "pkg/0.1@lasote/testing (test package): TESTING!!!" in client.out


def test_create_skip_test_package():
    # Skip the test package stage if explicitly disabled with --test-folder=None
    # https://github.com/conan-io/conan/issues/2355
    client = TestClient()
    client.save({"conanfile.py": GenConanfile().with_name("pkg").with_version("0.1"),
                 "test_package/conanfile.py":
                     GenConanfile().with_test('self.output.info("TESTING!!!")')})
    client.run("create . --user=lasote --channel=testing --test-folder=\"\"")
    assert "pkg/0.1@lasote/testing: Generating the package" in client.out
    assert "TESTING!!!" not in client.out


def test_create_package_requires():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile()})
    client.run("create . --name=dep --version=0.1 --user=user --channel=channel")
    client.run("create . --name=other --version=1.0 --user=user --channel=channel")

    conanfile = GenConanfile().with_require("dep/0.1@user/channel")
    test_conanfile = textwrap.dedent("""
    from conan import ConanFile
    class MyPkg(ConanFile):
        requires = "other/1.0@user/channel"
        def requirements(self):
            self.requires(self.tested_reference_str)
        def build(self):
            for r in self.requires.values():
                self.output.info("build() Requires: %s" % str(r.ref))
            import os
            for dep in self.dependencies.host.values():
                self.output.info("build() cpp_info dep: %s" % dep)

        def test(self):
            pass
    """)

    client.save({"conanfile.py": conanfile,
                 "test_package/conanfile.py": test_conanfile})

    client.run("create . --name=pkg --version=0.1 --user=lasote --channel=testing -vv")

    assert "pkg/0.1@lasote/testing (test package): build() " \
           "Requires: other/1.0@user/channel" in client.out
    assert "pkg/0.1@lasote/testing (test package): build() " \
           "Requires: pkg/0.1@lasote/testing" in client.out
    assert "pkg/0.1@lasote/testing (test package): build() cpp_info dep: other" in client.out
    assert "pkg/0.1@lasote/testing (test package): build() cpp_info dep: dep" in client.out
    assert "pkg/0.1@lasote/testing (test package): build() cpp_info dep: pkg" in client.out

    # Check that the additional info shows up when creating pkg
    # Graph info, package creation, and test package graph info
    assert client.out.count("requires: dep/0.1@user/channel") == 3


def test_package_folder_build_error():
    """
    Check package folder is not created if the build step fails
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class MyPkg(ConanFile):

            def build(self):
                raise Exception("Build error")
        """)
    client.save({"conanfile.py": conanfile})

    ref = RecipeReference("pkg", "0.1", "danimtb", "testing")
    client.run("create . --name=pkg --version=0.1 --user=danimtb --channel=testing",
               assert_error=True)

    assert "Build error" in client.out
    pref = client.get_latest_package_reference(ref, NO_SETTINGS_PACKAGE_ID)
    assert pref is None


def test_create_with_name_and_version():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile()})
    client.run('create . --name=lib --version=1.0')
    assert "lib/1.0: Created package revision" in client.out


def test_create_with_only_user_channel():
    """This should be the recommended way and only from Conan 2.0"""
    client = TestClient()
    client.save({"conanfile.py": GenConanfile().with_name("lib").with_version("1.0")})
    client.run('create . --user=user --channel=channel')
    assert "lib/1.0@user/channel: Created package revision" in client.out

    client.run('create . --user=user --channel=channel')
    assert "lib/1.0@user/channel: Created package revision" in client.out


def test_requires_without_user_channel():
    client = TestClient()
    conanfile = textwrap.dedent('''
        from conan import ConanFile

        class HelloConan(ConanFile):
            name = "hellobar"
            version = "0.1"

            def package_info(self):
                self.output.warning("Hello, I'm hellobar")
        ''')

    client.save({"conanfile.py": conanfile})
    client.run("create .")

    client.save({"conanfile.py": GenConanfile().with_require("hellobar/0.1")})
    client.run("create . --name=consumer --version=1.0")
    assert "hellobar/0.1: WARN: Hello, I'm hellobar" in client.out
    assert "consumer/1.0: Created package revision" in client.out


def test_conaninfo_contents_without_user_channel():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile().with_name("hello").with_version("0.1")})
    client.run("create .")
    client.save({"conanfile.py": GenConanfile().with_name("bye").with_version("0.1")
                .with_require("hello/0.1")})
    client.run("create .")

    package_folder = client.created_layout().package()

    conaninfo = load(os.path.join(package_folder, "conaninfo.txt"))
    # The user and channel nor None nor "_/" appears in the conaninfo
    assert "None" not in conaninfo
    assert "_/" not in conaninfo
    assert "/_" not in conaninfo
    assert "[requires]\nhello/0.1\n" in conaninfo


def test_components_json_output():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class MyTest(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "build_type"

            def package_info(self):
                self.cpp_info.components["pkg1"].libs = ["libpkg1"]
                self.cpp_info.components["pkg2"].libs = ["libpkg2"]
                self.cpp_info.components["pkg2"].requires = ["pkg1"]
        """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --format=json")
    data = json.loads(client.stdout)
    cpp_info_data = data["graph"]["nodes"]["1"]["cpp_info"]
    assert "libpkg1" in cpp_info_data["pkg1"]["libs"]
    assert cpp_info_data["pkg1"]["requires"] == []
    assert "libpkg2" in cpp_info_data["pkg2"]["libs"]
    assert cpp_info_data["pkg2"]["requires"] == ["pkg1"]


def test_lockfile_input_not_specified():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile().with_name("foo").with_version("1.0")})
    client.run("lock create . --lockfile-out locks/conan.lock")
    client.run("create . --lockfile-out locks/conan.lock")
    assert "Generated lockfile:" in client.out


def test_create_build_missing():
    """ test the --build=missing:pattern syntax
    """
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile("dep", "1.0").with_settings("os"),
            "pkg/conanfile.py": GenConanfile("pkg", "1.0").with_settings("os")
           .with_requires("dep/1.0")})
    c.run("create dep -s os=Windows")

    # Wrong pattern will not build it
    c.run("create pkg -s os=Windows --build=missing:kk", assert_error=True)
    assert "ERROR: Missing prebuilt package for 'pkg/1.0'" in c.out

    # Pattern missing * will not build it
    c.run("create pkg -s os=Windows --build=missing:pkg", assert_error=True)
    assert "ERROR: Missing prebuilt package for 'pkg/1.0'" in c.out

    # Correct pattern pkg* will build it
    c.run("create pkg -s os=Windows --build=missing:pkg*")
    c.assert_listed_binary({"pkg/1.0": ("90887fdbe22295dfbe41afe0a45f960c6a72b650", "Build")})

    # Now anything that is not an explicit --build=pkg* will avoid rebuilding
    c.run("create pkg -s os=Windows --build=missing:kk")
    c.assert_listed_binary({"pkg/1.0": ("90887fdbe22295dfbe41afe0a45f960c6a72b650", "Cache")})
    assert "Calling build()" not in c.out

    # but dependency without binary will fail, even if right pkg* pattern
    c.run("create pkg -s os=Linux --build=missing:pkg*", assert_error=True)
    c.assert_listed_binary({"pkg/1.0": ("4c0c198b627f9af3e038af4da5e6b3ae205c2435", "Build")})
    c.assert_listed_binary({"dep/1.0": ("9a4eb3c8701508aa9458b1a73d0633783ecc2270", "Missing")})
    assert "ERROR: Missing prebuilt package for 'dep/1.0'" in c.out

    # The & placeholder also works
    c.run("create pkg -s os=Linux --build=missing:&", assert_error=True)
    c.assert_listed_binary({"pkg/1.0": ("4c0c198b627f9af3e038af4da5e6b3ae205c2435", "Build")})
    c.assert_listed_binary({"dep/1.0": ("9a4eb3c8701508aa9458b1a73d0633783ecc2270", "Missing")})
    assert "ERROR: Missing prebuilt package for 'dep/1.0'" in c.out


def test_create_no_user_channel():
    """ test the --build=pattern and --build=missing:pattern syntax to build missing packages
     without user/channel
    """
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile(),
            "pkg/conanfile.py": GenConanfile("pkg", "1.0").with_requires("dep1/0.1", "dep2/0.1@user",
                                                                         "dep3/0.1@user/channel")})
    c.run("export dep --name=dep1 --version=0.1")
    c.run("export dep --name=dep2 --version=0.1 --user=user")
    c.run("export dep --name=dep3 --version=0.1 --user=user --channel=channel")

    # First test the ``--build=missing:pattern``
    c.run("create pkg --build=missing:*@", assert_error=True)
    c.assert_listed_binary({"dep1/0.1": (NO_SETTINGS_PACKAGE_ID, "Build"),
                            "dep2/0.1": (NO_SETTINGS_PACKAGE_ID, "Missing"),
                            "dep3/0.1": (NO_SETTINGS_PACKAGE_ID, "Missing")})
    c.run("create pkg --build=missing:!*@", assert_error=True)
    c.assert_listed_binary({"dep1/0.1": (NO_SETTINGS_PACKAGE_ID, "Missing"),
                            "dep2/0.1": (NO_SETTINGS_PACKAGE_ID, "Build"),
                            "dep3/0.1": (NO_SETTINGS_PACKAGE_ID, "Build")})

    # Now lets make sure they exist
    c.run("create pkg --build=missing")

    # Now test the --build=pattern
    c.run("create pkg --build=*@")
    c.assert_listed_binary({"dep1/0.1": (NO_SETTINGS_PACKAGE_ID, "Build"),
                            "dep2/0.1": (NO_SETTINGS_PACKAGE_ID, "Cache"),
                            "dep3/0.1": (NO_SETTINGS_PACKAGE_ID, "Cache")})
    # The --build=* needs to be said: "build all except those that have user/channel
    c.run("create pkg --build=* --build=!*@")
    c.assert_listed_binary({"dep1/0.1": (NO_SETTINGS_PACKAGE_ID, "Cache"),
                            "dep2/0.1": (NO_SETTINGS_PACKAGE_ID, "Build"),
                            "dep3/0.1": (NO_SETTINGS_PACKAGE_ID, "Build")})


def test_create_build_missing_negation():
    tc = TestClient(light=True)
    tc.save({"dep/conanfile.py": GenConanfile("dep", "1.0"),
             "lib/conanfile.py": GenConanfile("lib", "1.0").with_requires("dep/1.0"),
             "pkg/conanfile.py": GenConanfile("pkg", "1.0").with_requires("lib/1.0")})

    tc.run("export dep")
    tc.run("export lib")
    tc.run("create pkg --build=missing:~dep/*", assert_error=True)

    tc.assert_listed_binary({"pkg/1.0": ("a72376edfbbdaf97c8608b5fda53cadebac46a20", "Build"),
                             "lib/1.0": ("abfcc78fa8242cabcd1e3d92896aa24808c789a3", "Build"),
                             "dep/1.0": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Missing")})

    tc.run("create pkg --build=missing:~dep/* --build=missing:~lib/*",
           assert_error=True)

    tc.assert_listed_binary({"pkg/1.0": ("a72376edfbbdaf97c8608b5fda53cadebac46a20", "Build"),
                             "lib/1.0": ("abfcc78fa8242cabcd1e3d92896aa24808c789a3", "Missing"),
                             "dep/1.0": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Missing")})


def test_create_format_json():
    """
    Tests the ``conan create . -f json`` result

    The result should be something like:

    {
        'graph': {
            'nodes': [
                {'ref': '',  # consumer
                 'recipe': 'Virtual',
                 ....
                },
                {'ref': 'hello/0.1#18d5440ae45afc4c36139a160ac071c7',
                 'dependencies': {'1': {'ref': 'hello/0.1', 'visible': 'True', ...}},
                 ....
                },
                {'ref': 'pkg/0.2#44a1a27ac2ea1fbcf434a05c4d57388d',
                 ....
                }
            ],
            'root': {'0': 'None'}
        }
    }
    """
    client = TestClient()
    profile_build = textwrap.dedent("""\
    [settings]
    arch=x86_64
    build_type=Release
    compiler=gcc
    compiler.libcxx=libstdc++
    compiler.version=12
    os=Linux
    [conf]
    user.first:value="my value"
    user.second:value=["my value"]
    user.second:value+=["other value"]
    [buildenv]
    VAR1=myvalue1
    """)
    profile_host = textwrap.dedent("""\
    [settings]
    arch=x86
    build_type=Debug
    compiler=gcc
    compiler.libcxx=libstdc++
    compiler.version=12
    os=Linux
    """)
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class MyTest(ConanFile):
            name = "pkg"
            version = "0.2"
            settings = "build_type", "compiler"
            author = "John Doe"
            license = "MIT"
            url = "https://foo.bar.baz"
            homepage = "https://foo.bar.site"
            topics = "foo", "bar", "qux"
            provides = "libjpeg", "libjpg"
            deprecated = "other-pkg"
            options = {"shared": [True, False], "fPIC": [True, False]}
            default_options = {"shared": False, "fPIC": True}
        """)
    client.save({"conanfile.py": conanfile,
                 "host": profile_host, "build": profile_build})
    client.run("create . -pr:h host -pr:b build")
    client.save({"conanfile.py": GenConanfile().with_name("hello").with_version("0.1")
                .with_require("pkg/0.2"),
                 "host": profile_host, "build": profile_build}, clean_first=True)
    client.run("create . -f json -pr:h host -pr:b build")
    info = json.loads(client.stdout)
    nodes = info["graph"]['nodes']
    consumer_ref = 'conanfile'
    hello_pkg_ref = 'hello/0.1#18d5440ae45afc4c36139a160ac071c7'
    pkg_pkg_ref = 'pkg/0.2#db78b8d06a78af5c3ac56706f133098d'
    consumer_info = hello_pkg_info = pkg_pkg_info = None

    for n in nodes.values():
        ref = n["ref"]
        if ref == consumer_ref:
            consumer_info = n
        elif ref == hello_pkg_ref:
            hello_pkg_info = n
        else:
            assert ref == pkg_pkg_ref
            pkg_pkg_info = n

    # Consumer information
    assert consumer_info["recipe"] == "Cli"
    assert consumer_info["package_id"] is None
    assert consumer_info["prev"] is None
    assert consumer_info["options"] == {}
    assert consumer_info["settings"] == {'arch': 'x86', 'build_type': 'Debug', 'compiler': 'gcc',
                                         'compiler.libcxx': 'libstdc++', 'compiler.version': '12',
                                         'os': 'Linux'}
    consumer_deps = {
        '1': {'ref': 'hello/0.1', 'run': False, 'libs': True, 'skip': False,
              'test': False, 'force': False, 'direct': True, 'build': False,
              'transitive_headers': None, 'transitive_libs': None, 'headers': True,
              'package_id_mode': None, 'visible': True, 'require': 'hello/0.1'},
        '2': {'ref': 'pkg/0.2', 'run': False, 'libs': True, 'skip': False, 'test': False,
              'force': False, 'direct': False, 'build': False, 'transitive_headers': None,
              'transitive_libs': None, 'headers': True, 'package_id_mode': None,
              'visible': True, 'require': 'pkg/0.2'}
    }
    assert consumer_info["dependencies"] == consumer_deps
    # hello/0.1 pkg information
    assert hello_pkg_info["package_id"] == "8eba237c0fb239fcb7daa47979ab99258eaaa7d1"
    assert hello_pkg_info["prev"] == "d95380a07c35273509dfc36b26f6cec1"
    assert hello_pkg_info["settings"] == {}
    assert hello_pkg_info["options"] == {}
    hello_pkg_info_deps = {
        "2": {
            "ref": "pkg/0.2", "run": False, "libs": True, "skip": False, "test": False,
            "force": False, "direct": True, "build": False, "transitive_headers": None,
            "transitive_libs": None, "headers": True, "package_id_mode": "semver_mode",
            "visible": True, 'require': 'pkg/0.2'
        }
    }
    assert hello_pkg_info["dependencies"] == hello_pkg_info_deps
    # pkg/0.2 pkg information
    assert pkg_pkg_info["package_id"] == "fb1439470288b15b2da269ed97b1a5f2f5d1f766"
    assert pkg_pkg_info["prev"] == "6949b0f89941d2a5994f9e6e4a89a331"
    assert pkg_pkg_info["author"] == 'John Doe'
    assert pkg_pkg_info["settings"] == {'build_type': 'Debug', 'compiler': 'gcc',
                                        'compiler.libcxx': 'libstdc++', 'compiler.version': '12'}
    assert pkg_pkg_info["options"] == {'fPIC': 'True', 'shared': 'False'}
    assert pkg_pkg_info["dependencies"] == {}


def test_create_format_json_and_deps_cpp_info():
    """
    Tests the ``conan create . -f json`` result, but ``cpp_info`` object only.

    The goal is to get something like:

    ```
    { ....
    'cpp_info': {'cmp1': {'bindirs': None,
                          'builddirs': None,
                          'cflags': None,
                          'cxxflags': None,
                          'defines': None,
                          'exelinkflags': None,
                          'frameworkdirs': None,
                          'frameworks': None,
                          'includedirs': None,
                          'libdirs': None,
                          'libs': ['libcmp1'],
                          'objects': None,
                          'properties': {'pkg_config_aliases': ['compo1_alias'],
                                         'pkg_config_name': 'compo1'},
                          'dependencies': None,
                          'resdirs': None,
                          'sharedlinkflags': None,
                          'srcdirs': None,
                          'sysroot': '/another/sysroot',
                          'system_libs': None},
                 'root': {'bindirs': ['bin'],
                          'builddirs': [],
                          'cflags': ['pkg_a_c_flag'],
                          'cxxflags': ['pkg_a_cxx_flag'],
                          'defines': ['pkg_onedefinition',
                                      'pkg_twodefinition'],
                          'exelinkflags': ['pkg_exe_link_flag'],
                          'frameworkdirs': ['framework/path/pkg'],
                          'frameworks': ['pkg_oneframework',
                                         'pkg_twoframework'],
                          'includedirs': ['path/includes/pkg',
                                          'include/path/pkg'],
                          'libdirs': ['lib/path/pkg'],
                          'libs': ['pkg'],
                          'objects': None,
                          'properties': {'pkg_config_aliases': ['pkg_alias1',
                                                                'pkg_alias2'],
                                         'pkg_config_name': 'pkg_other_name'},
                          'dependencies': None,
                          'resdirs': ['/path '
                                      'with '
                                      'spaces/.conan2/p/d15a235e212166d9/p/res'],
                          'sharedlinkflags': ['pkg_shared_link_flag'],
                          'srcdirs': None,
                          'sysroot': '/path/to/folder/pkg',
                          'system_libs': ['pkg_onesystemlib',
                                          'pkg_twosystemlib']}
    }}
    ```
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class MyTest(ConanFile):
            name = "pkg"
            version = "0.2"

            def package_info(self):
                self.cpp_info.libs = ["pkg"]
                self.cpp_info.includedirs = ["path/includes/pkg", "other/include/path/pkg"]
                self.cpp_info.libdirs = ["one/lib/path/pkg"]
                self.cpp_info.defines = ["pkg_onedefinition", "pkg_twodefinition"]
                self.cpp_info.cflags = ["pkg_a_c_flag"]
                self.cpp_info.cxxflags = ["pkg_a_cxx_flag"]
                self.cpp_info.sharedlinkflags = ["pkg_shared_link_flag"]
                self.cpp_info.exelinkflags = ["pkg_exe_link_flag"]
                self.cpp_info.sysroot = "/path/to/folder/pkg"
                self.cpp_info.frameworks = ["pkg_oneframework", "pkg_twoframework"]
                self.cpp_info.system_libs = ["pkg_onesystemlib", "pkg_twosystemlib"]
                self.cpp_info.frameworkdirs = ["one/framework/path/pkg"]
                self.cpp_info.set_property("pkg_config_name", "pkg_other_name")
                self.cpp_info.set_property("pkg_config_aliases", ["pkg_alias1", "pkg_alias2"])
                self.cpp_info.components["cmp1"].libs = ["libcmp1"]
                self.cpp_info.components["cmp1"].set_property("pkg_config_name", "compo1")
                self.cpp_info.components["cmp1"].set_property("pkg_config_aliases", ["compo1_alias"])
                self.cpp_info.components["cmp1"].sysroot = "/another/sysroot"
        """)
    client.save({"conanfile.py": conanfile})
    client.run("create .")
    client.save({"conanfile.py": GenConanfile().with_name("hello").with_version("0.1")
                .with_require("pkg/0.2")}, clean_first=True)
    client.run("create . -f json")
    info = json.loads(client.stdout)
    nodes = info["graph"]["nodes"]
    hello_pkg_ref = 'hello/0.1#18d5440ae45afc4c36139a160ac071c7'
    pkg_pkg_ref = 'pkg/0.2#926714b5fb0a994f47ec37e071eba1da'
    hello_cpp_info = pkg_cpp_info = None
    for n in nodes.values():
        ref = n["ref"]
        if ref == hello_pkg_ref:
            assert n['binary'] == "Build"
            hello_cpp_info = n['cpp_info']
        elif ref == pkg_pkg_ref:
            assert n['binary'] == "Cache"
            pkg_cpp_info = n['cpp_info']

    assert hello_cpp_info and pkg_cpp_info
    # hello/0.1 cpp_info
    assert hello_cpp_info['root']["libs"] is None
    assert len(hello_cpp_info['root']["bindirs"]) == 1
    assert len(hello_cpp_info['root']["libdirs"]) == 1
    assert hello_cpp_info['root']["sysroot"] is None
    assert hello_cpp_info['root']["properties"] is None
    # pkg/0.2 cpp_info
    # root info
    assert pkg_cpp_info['root']["libs"] == ['pkg']
    assert len(pkg_cpp_info['root']["bindirs"]) == 1
    assert len(pkg_cpp_info['root']["libdirs"]) == 1
    assert pkg_cpp_info['root']["sysroot"] == '/path/to/folder/pkg'
    assert pkg_cpp_info['root']["system_libs"] == ['pkg_onesystemlib', 'pkg_twosystemlib']
    assert pkg_cpp_info['root']['cflags'] == ['pkg_a_c_flag']
    assert pkg_cpp_info['root']['cxxflags'] == ['pkg_a_cxx_flag']
    assert pkg_cpp_info['root']['defines'] == ['pkg_onedefinition', 'pkg_twodefinition']
    assert pkg_cpp_info['root']["properties"] == {'pkg_config_aliases': ['pkg_alias1', 'pkg_alias2'],
                                                  'pkg_config_name': 'pkg_other_name'}
    # component info
    assert pkg_cpp_info['cmp1']["libs"] == ['libcmp1']
    assert pkg_cpp_info['cmp1']["bindirs"][0].endswith("bin")  # Abs path /bin
    assert pkg_cpp_info['cmp1']["libdirs"][0].endswith("lib")  # Abs path /lib
    assert pkg_cpp_info['cmp1']["sysroot"] == "/another/sysroot"
    assert pkg_cpp_info['cmp1']["properties"] == {'pkg_config_aliases': ['compo1_alias'],
                                                  'pkg_config_name': 'compo1'}


def test_default_framework_dirs():
    conanfile = textwrap.dedent("""
    from conan import ConanFile


    class LibConan(ConanFile):
        name = "lib"
        version = "1.0"

        def package_info(self):
            self.output.warning("FRAMEWORKS: {}".format(self.cpp_info.frameworkdirs))""")
    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run("create .")
    assert "FRAMEWORKS: []" in client.out


def test_default_framework_dirs_with_layout():
    conanfile = textwrap.dedent("""
    from conan import ConanFile


    class LibConan(ConanFile):
        name = "lib"
        version = "1.0"

        def layout(self):
            pass

        def package_info(self):
            self.output.warning("FRAMEWORKS: {}".format(self.cpp_info.frameworkdirs))""")
    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run("create .")
    assert "FRAMEWORKS: []" in client.out


def test_defaults_in_components():
    """In Conan 2, declaring or not the layout has no influence in how cpp_info behaves. It was
       only 1.X"""
    lib_conan_file = textwrap.dedent("""
    from conan import ConanFile

    class LibConan(ConanFile):
        name = "lib"
        version = "1.0"

        def layout(self):
            pass

        def package_info(self):
            self.cpp_info.components["foo"].libs = ["foolib"]

    """)
    client = TestClient()
    client.save({"conanfile.py": lib_conan_file})
    client.run("create . ")

    consumer_conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Consumer(ConanFile):
            name = "consumer"
            version = "1.0"
            requires = "lib/1.0"

            def layout(self):
                pass

            def generate(self):
                cppinfo = self.dependencies["lib"].cpp_info
                components = cppinfo.components
                self.output.warning("BINDIRS: {}".format(cppinfo.bindirs))
                self.output.warning("LIBDIRS: {}".format(cppinfo.libdirs))
                self.output.warning("INCLUDEDIRS: {}".format(cppinfo.includedirs))
                self.output.warning("RESDIRS: {}".format(cppinfo.resdirs))
                self.output.warning("FOO LIBDIRS: {}".format(components["foo"].libdirs))
                self.output.warning("FOO INCLUDEDIRS: {}".format(components["foo"].includedirs))
                self.output.warning("FOO RESDIRS: {}".format(components["foo"].resdirs))

        """)

    client.save({"conanfile.py": consumer_conanfile})
    client.run("create . ")

    # The paths are absolute and the components have defaults
    # ".+" Check that there is a path, not only "lib"
    assert re.search(r"BINDIRS: \['.+bin']", client.out)
    assert re.search(r"LIBDIRS: \['.+lib']", client.out)
    assert re.search(r"INCLUDEDIRS: \['.+include']", client.out)
    assert "WARN: RES DIRS: []"
    assert re.search(r"WARN: FOO LIBDIRS: \['.+lib']", client.out)
    assert re.search(r"WARN: FOO INCLUDEDIRS: \['.+include']", client.out)
    assert "WARN: FOO RESDIRS: []" in client.out

    # The paths are absolute and the components have defaults
    # ".+" Check that there is a path, not only "lib"
    assert re.search(r"BINDIRS: \['.+bin']", client.out)
    assert re.search(r"LIBDIRS: \['.+lib']", client.out)
    assert re.search(r"INCLUDEDIRS: \['.+include']", client.out)
    assert "WARN: RES DIRS: []"
    assert bool(re.search(r"WARN: FOO LIBDIRS: \['.+lib']", client.out))
    assert bool(re.search(r"WARN: FOO INCLUDEDIRS: \['.+include']", client.out))
    assert "WARN: FOO RESDIRS: []" in client.out


def test_name_never():
    """ check that a package can be named equal to a build policy --build=never,
    because --build are now patterns
    Close https://github.com/conan-io/conan/issues/12430
    """
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("never", "0.1")})
    c.run("create .")
    assert "never/0.1: Created package" in c.out


def test_create_both_host_build_require():
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("protobuf", "0.1").with_settings("build_type"),
            "test_package/conanfile.py": GenConanfile().with_build_requires("protobuf/0.1")
                                                       .with_test("pass")})
    c.run("create . -s:b build_type=Release -s:h build_type=Debug", assert_error=True)
    # The main "host" Debug binary will be correctly build
    c.assert_listed_binary({"protobuf/0.1": ("9e186f6d94c008b544af1569d1a6368d8339efc5", "Build")})
    # But test_package will fail because of the missing "tool_require" in Release
    c.assert_listed_binary({"protobuf/0.1": ("efa83b160a55b033c4ea706ddb980cd708e3ba1b", "Missing")},
                           build=True, test_package=True)

    c.run("remove * -c")  # make sure that previous binary is removed
    c.run("create . -s:b build_type=Release -s:h build_type=Debug --build-test=missing")
    c.assert_listed_binary({"protobuf/0.1": ("9e186f6d94c008b544af1569d1a6368d8339efc5", "Build")})
    # it used to fail, now it works and builds the test_package "tools_requires" in Release
    c.assert_listed_binary({"protobuf/0.1": ("9e186f6d94c008b544af1569d1a6368d8339efc5", "Cache")},
                           test_package=True)
    c.assert_listed_binary({"protobuf/0.1": ("efa83b160a55b033c4ea706ddb980cd708e3ba1b", "Build")},
                           build=True, test_package=True)

    # we can be more explicit about the current package only with "missing:protobuf/*"
    c.run("remove * -c")  # make sure that previous binary is removed
    c.run("create . -s:b build_type=Release -s:h build_type=Debug --build-test=missing:protobuf/*")
    c.assert_listed_binary({"protobuf/0.1": ("9e186f6d94c008b544af1569d1a6368d8339efc5", "Build")})
    # it used to fail, now it works and builds the test_package "tools_requires" in Release
    c.assert_listed_binary({"protobuf/0.1": ("efa83b160a55b033c4ea706ddb980cd708e3ba1b", "Build")},
                           build=True, test_package=True)


def test_python_requires_json_format():
    """Check python requires does not crash when calling conan create . --format=json
    See https://github.com/conan-io/conan/issues/14577"""
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("pyreq", "1.0")
           .with_package_type("python-require")})
    c.run("create . --format=json", redirect_stdout="output.json")
    data = json.loads(load(os.path.join(c.current_folder, "output.json")))
    # There's a graph and the python requires is there
    assert len(data["graph"]["nodes"]["0"]["python_requires"]) == 1


def test_python_requires_with_test_package():
    c = TestClient()
    # Code comes from the docs
    conanfile = textwrap.dedent("""
    from conan import ConanFile

    def mynumber():
        return 42

    class PyReq(ConanFile):
        name = "pyreq"
        version = "1.0"
        package_type = "python-require"
    """)
    test_conanfile = textwrap.dedent("""
    from conan import ConanFile

    class Tool(ConanFile):
        def test(self):
            pyreq = self.python_requires["pyreq"].module
            mynumber = pyreq.mynumber()
            self.output.info("{}!!!".format(mynumber))
    """)
    c.save({"conanfile.py": conanfile, "test_package/conanfile.py": test_conanfile})
    c.run("create .")
    # Ensure that creating a deps graph does not break the testing
    assert "pyreq/1.0 (test package): 42!!!" in c.out


def test_create_test_package_only_build():
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("pkg", "0.1"),
            "test_package/conanfile.py": GenConanfile().with_test("self.output.info('TEST1!!!')"),
            "test_package2/conanfile.py": GenConanfile().with_test("self.output.info('TEST2!!!')")})
    # As it doesn't exist, it builds and test it
    c.run("create . -tm")
    assert "Testing the package" in c.out
    assert "TEST1!!!" in c.out
    # this will not create the binary, so it won't test it
    c.run("create . --build=missing --test-missing")
    assert "Testing the package" not in c.out
    assert "TEST" not in c.out
    c.run("create . -tf=test_package2 -tm")
    assert "Testing the package" in c.out
    assert "TEST2!!!" in c.out
    assert "TEST1!!!" not in c.out
    c.run("create . -tf=test_package2 --build=missing --test-missing")
    assert "Testing the package" not in c.out
    assert "TEST2!!!" not in c.out
    assert "TEST1!!!" not in c.out

    # error
    c.run("create . -tm -tf=", assert_error=True)
    assert '--test-folder="" is incompatible with --test-missing' in c.out


def test_create_test_package_only_build_python_require():
    c = TestClient()
    test = textwrap.dedent("""
        from conan import ConanFile

        class Tool(ConanFile):
            python_requires = "tested_reference_str"
            def test(self):
                self.output.info("TEST!!!!")
        """)
    c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_package_type("python-require"),
            "test_package/conanfile.py": test})
    c.run("create .")
    assert "Testing the package" in c.out
    assert "pkg/0.1 (test package): TEST!!!" in c.out
    c.run("create . -tm")
    assert "Testing the package" in c.out
    assert "pkg/0.1 (test package): TEST!!!" in c.out
    c.run("create . -tm --build=missing")
    assert "Testing the package" in c.out
    assert "pkg/0.1 (test package): TEST!!!" in c.out


@pytest.mark.parametrize("command", ["create", "install"])
@pytest.mark.parametrize("out_file", [False, True])
def test_create_build_fail_generate_outfile(command, out_file):
    c = TestClient()
    c.save({"pkga/conanfile.py": GenConanfile("pkga", "0.1"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "0.1").with_requires("pkga/0.1"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "0.1")
           .with_requires("pkgb/0.1")
           .with_package("raise Exception('myerror')"),
            "pkgd/conanfile.py": GenConanfile("pkgd", "0.1").with_requires("pkgc/0.1")
                                                            .with_settings("build_type")
                                                            .with_generator("CMakeDeps"),
            })
    c.run("export pkga")
    c.run("export pkgb")
    c.run("export pkgc")
    if out_file:
        error = c.run(f"{command} pkgd --build=missing --format=json --out-file=graph.json",
                      assert_error=True)
    else:
        error = c.run(f"{command} pkgd --build=missing --format=json", assert_error=True,
                      redirect_stdout="graph.json")
    assert error == ERROR_GENERAL
    assert "pkgc/0.1: Error in package() method, line 8" in c.out
    graph = json.loads(c.load("graph.json"))
    nodeid = "1" if command == "create" else "0"
    assert graph["graph"]["nodes"][nodeid]["name"] == "pkgd"

    # We can construct a package list from it
    c.run("list -g=graph.json --graph-binaries=Build --format=json")
    pkglist = json.loads(c.stdout)
    # not built packages don't have revisions
    rrev = pkglist["Local Cache"]["pkgc/0.1"]["revisions"]["b7f74fa20b19f1daac67db49318b7197"]
    assert "revisions" not in rrev["packages"]["4a8d7d78a454700be1ab74b4a77fd7f36a44d122"]
    # built packages do have package revisions
    rrev = pkglist["Local Cache"]["pkgb/0.1"]["revisions"]["5b1ae5e3c1f718c0fd90d4dd8d9b57fb"]
    assert "revisions" in rrev["packages"]["47a5f20ec8fb480e1c5794462089b01a3548fdc5"]
    rrev = pkglist["Local Cache"]["pkga/0.1"]["revisions"]["57ece23aeb368b634896004ad579767a"]
    assert "revisions" in rrev["packages"]["da39a3ee5e6b4b0d3255bfef95601890afd80709"]



import textwrap

import pytest

from conan.test.assets.sources import gen_function_cpp
from conan.test.utils.tools import TestClient


@pytest.mark.tool("cmake")
def test_cmake_virtualenv(matrix_client):
    client = matrix_client

    cmakewrapper = textwrap.dedent(r"""
        from conan import ConanFile
        import os
        from conan.tools.files import save, chdir
        class Pkg(ConanFile):
            def package(self):
                with chdir(self, self.package_folder):
                    save(self, "cmake.bat", "@echo off\necho MYCMAKE WRAPPER!!\ncmake.exe %*")
                    save(self, "cmake.sh", 'echo MYCMAKE WRAPPER!!\ncmake "$@"')
                    os.chmod("cmake.sh", 0o777)

            def package_info(self):
                # Custom buildenv not defined by cpp_info
                self.buildenv_info.prepend_path("PATH", self.package_folder)
            """)
    consumer = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake
        class App(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            exports_sources = "CMakeLists.txt", "main.cpp"
            requires = "matrix/1.0"
            build_requires = "cmakewrapper/0.1"
            generators = "CMakeDeps", "CMakeToolchain", "VirtualBuildEnv"

            def build(self):
                cmake = CMake(self)
                if self.settings.os != "Windows":
                    cmake._cmake_program = "cmake.sh"  # VERY DIRTY HACK
                cmake.configure()
                cmake.build()
        """)

    cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.15)
        project(MyApp CXX)

        find_package(matrix)
        add_executable(app main.cpp)
        target_link_libraries(app matrix::matrix)
        """)

    client.save({"cmakewrapper/conanfile.py": cmakewrapper,
                 "consumer/conanfile.py": consumer,
                 "consumer/main.cpp": gen_function_cpp(name="main", includes=["matrix"],
                                                       calls=["matrix"]),
                 "consumer/CMakeLists.txt": cmakelists},
                clean_first=True)

    client.run("create cmakewrapper --name=cmakewrapper --version=0.1")
    client.run("create consumer --name=consumer --version=0.1")
    assert "MYCMAKE WRAPPER!!" in client.out
    assert "consumer/0.1: Created package" in client.out


@pytest.mark.tool("cmake")
def test_complete():
    client = TestClient()
    client.run("new cmake_lib -d name=myopenssl -d version=1.0")
    client.run("create . -o myopenssl/*:shared=True")
    client.run("create . -o myopenssl/*:shared=True -s build_type=Debug")

    mycmake_main = gen_function_cpp(name="main", msg="mycmake",
                                    includes=["myopenssl"], calls=["myopenssl"])
    mycmake_conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.cmake import CMake
        from conan.tools.files import copy
        class App(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            requires = "myopenssl/1.0"
            default_options = {"myopenssl:shared": True}
            generators = "CMakeDeps", "CMakeToolchain", "VirtualBuildEnv"
            exports_sources = "*"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()

            def package(self):
                src = str(self.settings.build_type) if self.settings.os == "Windows" else ""
                copy(self, "mycmake*", os.path.join(self.source_folder, src),
                     os.path.join(self.package_folder, "bin"))

            def package_info(self):
                self.cpp_info.bindirs = ["bin"]
        """)
    mycmake_cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.15)
        project(MyCmake CXX)

        find_package(myopenssl REQUIRED)
        add_executable(mycmake main.cpp)
        target_link_libraries(mycmake PRIVATE myopenssl::myopenssl)
        """)
    client.save({"conanfile.py": mycmake_conanfile,
                 "CMakeLists.txt": mycmake_cmakelists,
                 "main.cpp": mycmake_main}, clean_first=True)
    client.run("create . --name=mycmake --version=1.0", assert_error=True)
    assert "The usage of package names `myopenssl:shared` in options is deprecated, " \
           "use a pattern like `myopenssl/*:shared` instead" in client.out

    client.run("create . --name=mycmake --version=1.0 -o=:shared=True", assert_error=True)
    assert "Invalid empty package" in client.out

    # Fix the default options and repeat the create
    fixed_cf = mycmake_conanfile.replace('default_options = {"myopenssl:shared": True}',
                                         'default_options = {"myopenssl*:shared": True}')
    client.save({"conanfile.py": fixed_cf})
    client.run("create . --name=mycmake --version=1.0")

    mylib = textwrap.dedent(r"""
        from conan import ConanFile
        import os
        from conan.tools.cmake import CMake
        class Pkg(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            build_requires = "mycmake/1.0"
            requires = "myopenssl/1.0"
            default_options = {"myopenssl/*:shared": True}
            exports_sources = "CMakeLists.txt", "main.cpp"
            generators = "CMakeDeps", "CMakeToolchain"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
                self.run("mycmake")
                self.output.info("RUNNING MYAPP")
                if self.settings.os == "Windows":
                    self.run(os.sep.join([".", str(self.settings.build_type), "myapp"]),
                             env="conanrun")
                else:
                    self.run(os.sep.join([".", "myapp"]), env=["conanrun"])
            """)

    cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.15)
        project(MyApp CXX)

        find_package(myopenssl)
        add_executable(myapp main.cpp)
        target_link_libraries(myapp myopenssl::myopenssl)
        """)

    client.save({"conanfile.py": mylib,
                 "main.cpp": gen_function_cpp(name="main", msg="myapp", includes=["myopenssl"],
                                              calls=["myopenssl"]),
                 "CMakeLists.txt": cmakelists},
                clean_first=True)

    client.run("create . --name=myapp --version=0.1 -s:b build_type=Release -s:h build_type=Debug")
    first, last = str(client.out).split("RUNNING MYAPP")
    assert "mycmake: Release!" in first
    assert "myopenssl/1.0: Hello World Release!" in first

    assert "myapp: Debug!" in last
    assert "myopenssl/1.0: Hello World Debug!" in last
